import copy
import logging
from enum import Enum
from logging import getLogger
import cz_ossfs
import s3fs
from google.oauth2.credentials import Credentials
from pyarrow import fs
import string
import gcsfs

from clickzetta.bulkload import cz_table
from clickzetta.proto.generated import ingestion_pb2, file_format_type_pb2

MAX_NUM_ROWS_PER_FILE = 64 << 20
MAX_FILE_SIZE_IN_BYTES_PER_FILE = 256 << 20


class FileFormatType(Enum):
    TEXT = 'text'
    PARQUET = 'parquet'
    ORC = 'orc'
    AVRO = 'avro'
    CSV = 'csv'
    ARROW = 'arrow'
    HIVE_RESULT = 'hive_result'
    DUMMY = 'dummy'
    MEMORY = 'memory'
    ICEBERG = 'iceberg'


class CZOSSFileSystem(cz_ossfs.OSSFileSystem):
    def __init__(self, endpoint: string, secret: string, token: string, key: string, cz_feature: string = None):
        super(CZOSSFileSystem, self).__init__(endpoint=endpoint, secret=secret, token=token, key=key)
        self.cz_feature = cz_feature

    def exists(self, path, **kwargs):
        if self.cz_feature == 'bulkload':
            bucket_name, obj_name = super().split_path(path)

            connect_timeout = kwargs.get("connect_timeout", None)
            if not obj_name:
                return True

            if super()._call_oss(
                    "object_exists",
                    obj_name,
                    bucket=bucket_name,
                    timeout=connect_timeout,
            ):
                return True

            return False
        else:
            return super().exists(path, **kwargs)


class StagingConfig:
    def __init__(self, path: string, id: string, secret: string, token: string, endpoint: string, type: None):
        self.path = path
        self.id = id
        self.secret = secret
        self.token = token
        self.endpoint = endpoint
        self.type = type

    def create_file_io(self):
        if self.type == 'oss':
            return CZOSSFileSystem(endpoint=self.endpoint, secret=self.secret, token=self.token, key=self.id,
                                   cz_feature='bulkload')
        elif self.type == 's3':
            client_kwargs = {'region_name': self.endpoint}
            return s3fs.S3FileSystem(key=self.id, secret=self.secret, token=self.token, client_kwargs=client_kwargs)
        elif self.type == 'gcs':
            credentials = Credentials(token=self.token)
            return gcsfs.GCSFileSystem(project='clickzetta-connector', token=credentials)
        elif self.type == 'tos':
            raise NotImplementedError('TOS is not supported bulkload yet')
            # TODO wait for volces to support s3fs. Currently, it got an error as below:
            # TypeError: _create_client() got an unexpected keyword argument 's3'
            # import s3fs
            # client_kwargs = {'s3': {'addressing_style': 'virtual'}}
            # return s3fs.S3FileSystem(key=self.id, secret=self.secret, token=self.token,
            #                          endpoint_url='https://tos-s3-cn-guangzhou.volces.com', client_kwargs=client_kwargs)
        else:
            return fs.LocalFileSystem()


class BulkLoadConfig:
    def __init__(self, config: ingestion_pb2.BulkLoadStreamWriterConfig):
        self.config = config

    def get_staging_config(self, prefer_internal_endpoint=False):
        staging_path = self.config.staging_path
        staging_config = None
        if staging_path.oss_path.path != '':
            oss_path = staging_path.oss_path
            endpoint = oss_path.oss_endpoint
            if prefer_internal_endpoint:
                endpoint = oss_path.oss_internal_endpoint
            staging_config = StagingConfig(oss_path.path, oss_path.sts_ak_id, oss_path.sts_ak_secret,
                                           oss_path.sts_token, endpoint, 'oss')
        elif staging_path.cos_path.path != '':
            cos_path = staging_path.cos_path
            staging_config = StagingConfig(cos_path.path, cos_path.sts_ak_id, cos_path.sts_ak_secret,
                                           cos_path.sts_token,
                                           cos_path.cos_region, 'cos')
        elif staging_path.s3_path.path != '':
            s3_path = staging_path.s3_path
            staging_config = StagingConfig(s3_path.path, s3_path.sts_ak_id, s3_path.sts_ak_secret,
                                           s3_path.sts_token,
                                           s3_path.s3_region, 's3')
        elif staging_path.gcs_path.path != '':
            gcs_path = staging_path.gcs_path
            staging_config = StagingConfig(gcs_path.path, gcs_path.sts_ak_id, gcs_path.sts_ak_secret,
                                           gcs_path.sts_token,
                                           gcs_path.gcs_region, 'gcs')
        elif staging_path.tos_path.path != '':
            tos_path = staging_path.tos_path
            # replace the tos:// to https://tos-s3-cn-guangzhou.volces.com
            tos_path.path = tos_path.path.replace('tos://', 's3://')
            staging_config = StagingConfig(tos_path.path, tos_path.sts_ak_id, tos_path.sts_ak_secret,
                                           tos_path.sts_token,
                                           tos_path.tos_endpoint, 'tos')
        else:
            raise ValueError('Staging path is not set, please check the configuration')
        return staging_config

    def get_file_format(self):
        file_format = self.config.file_format
        if file_format == file_format_type_pb2.FileFormatType.TEXT:
            return FileFormatType.TEXT
        elif file_format == file_format_type_pb2.FileFormatType.PARQUET:
            return FileFormatType.PARQUET
        elif file_format == file_format_type_pb2.FileFormatType.ORC:
            return FileFormatType.ORC
        elif file_format == file_format_type_pb2.FileFormatType.AVRO:
            return FileFormatType.AVRO
        elif file_format == file_format_type_pb2.FileFormatType.CSV:
            return FileFormatType.CSV
        elif file_format == file_format_type_pb2.FileFormatType.ARROW:
            return FileFormatType.ARROW
        elif file_format == file_format_type_pb2.FileFormatType.HIVE_RESULT:
            return FileFormatType.HIVE_RESULT
        elif file_format == file_format_type_pb2.FileFormatType.DUMMY:
            return FileFormatType.DUMMY
        elif file_format == file_format_type_pb2.FileFormatType.MEMORY:
            return FileFormatType.MEMORY
        elif file_format == file_format_type_pb2.FileFormatType.ICEBERG:
            return FileFormatType.ICEBERG

    def get_max_rows_per_file(self):
        if self.config.max_num_rows_per_file > 0:
            return self.config.max_num_rows_per_file
        return MAX_NUM_ROWS_PER_FILE

    def get_max_file_size_per_file(self):
        if self.config.max_size_in_bytes_per_file > 0:
            return self.config.max_size_in_bytes_per_file
        return MAX_FILE_SIZE_IN_BYTES_PER_FILE


class BulkLoadOperation(Enum):
    APPEND = 1
    UPSERT = 2
    OVERWRITE = 3


class BulkLoadState(Enum):
    CREATED = 1
    SEALED = 2
    COMMIT_SUBMITTED = 3
    COMMIT_SUCCESS = 4
    COMMIT_FAILED = 5
    ABORTED = 6


class BulkLoadOptions:
    def __init__(self, operation: BulkLoadOperation, partition_specs: str, record_keys: list,
                 prefer_internal_endpoint=False) -> None:
        self.operation = operation
        self.partition_specs = partition_specs
        self.record_keys = record_keys
        self.prefer_internal_endpoint = prefer_internal_endpoint
        self._properties = {'operation': operation, 'partition_specs': partition_specs, 'record_keys': record_keys,
                            'prefer_internal_endpoint': prefer_internal_endpoint}


    def to_api_repr(self) -> dict:
        return copy.deepcopy(self._properties)


class BulkLoadMetaData:
    def __init__(self, instance_id: int, info: ingestion_pb2.BulkLoadStreamInfo):
        self.instance_id = instance_id
        self.info = info
        self.table = cz_table.CZTable(info.stream_schema, info.identifier.schema_name, info.identifier.table_name)

    def get_instance_id(self):
        return self.instance_id

    def get_workspace(self):
        return self.info.identifier.workspace

    def get_schema_name(self):
        return self.info.identifier.schema_name

    def get_table_name(self):
        return self.info.identifier.table_name

    def get_stream_id(self):
        return self.info.stream_id

    def get_table(self):
        return self.table

    def get_prefer_internal_endpoint(self):
        return self.info.prefer_internal_endpoint

    def get_operation(self):
        if self.info.operation == ingestion_pb2.BulkLoadStreamOperation.BL_APPEND:
            return BulkLoadOperation.APPEND
        elif self.info.operation == ingestion_pb2.BulkLoadStreamOperation.BL_OVERWRITE:
            return BulkLoadOperation.OVERWRITE
        elif self.info.operation == ingestion_pb2.BulkLoadStreamOperation.BL_UPSERT:
            return BulkLoadOperation.UPSERT

    def get_state(self):
        if self.info.stream_state == ingestion_pb2.BulkLoadStreamState.BL_CREATED:
            return BulkLoadState.CREATED
        elif self.info.stream_state == ingestion_pb2.BulkLoadStreamState.BL_SEALED:
            return BulkLoadState.SEALED
        elif self.info.stream_state == ingestion_pb2.BulkLoadStreamState.BL_COMMIT_SUBMITTED:
            return BulkLoadState.COMMIT_SUBMITTED
        elif self.info.stream_state == ingestion_pb2.BulkLoadStreamState.BL_COMMIT_SUCCESS:
            return BulkLoadState.COMMIT_SUCCESS
        elif self.info.stream_state == ingestion_pb2.BulkLoadStreamState.BL_COMMIT_FAILED:
            return BulkLoadState.COMMIT_FAILED
        elif self.info.stream_state == ingestion_pb2.BulkLoadStreamState.BL_ABORTED:
            return BulkLoadState.ABORTED

    def get_sql_error_msg(self):
        return self.info.sql_error_msg

    def get_partition_specs(self):
        return self.info.partition_spec

    def get_record_keys(self):
        record_keys = []
        for key in self.info.record_keys:
            record_keys.append(key)

        return record_keys


class BulkLoadCommitOptions:
    def __init__(self, workspace: string, vc: string):
        self.workspace = workspace
        self.vc = vc


class BulkLoadCommitMode(Enum):
    COMMIT_STREAM = 1
    ABORT_STREAM = 2



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: expression.proto
# Protobuf Python Version: 5.28.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    1,
    '',
    'expression.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import data_type_pb2 as data__type__pb2
from . import property_pb2 as property__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x10\x65xpression.proto\x12\x08\x63z.proto\x1a\x0f\x64\x61ta_type.proto\x1a\x0eproperty.proto\"1\n\x0fIntervalDayTime\x12\x0f\n\x07seconds\x18\x01 \x01(\x03\x12\r\n\x05nanos\x18\x02 \x01(\x05\"2\n\nArrayValue\x12$\n\x08\x65lements\x18\x01 \x03(\x0b\x32\x12.cz.proto.Constant\"P\n\x08MapValue\x12 \n\x04keys\x18\x01 \x03(\x0b\x32\x12.cz.proto.Constant\x12\"\n\x06values\x18\x02 \x03(\x0b\x32\x12.cz.proto.Constant\"1\n\x0bStructValue\x12\"\n\x06\x66ields\x18\x01 \x03(\x0b\x32\x12.cz.proto.Constant\"\xe6\x03\n\x08\x43onstant\x12\x0e\n\x04null\x18\x01 \x01(\x08H\x00\x12\x11\n\x07tinyint\x18\x02 \x01(\x05H\x00\x12\x12\n\x08smallInt\x18\x03 \x01(\x05H\x00\x12\r\n\x03int\x18\x04 \x01(\x05H\x00\x12\x10\n\x06\x62igint\x18\x05 \x01(\x03H\x00\x12\x0f\n\x05\x66loat\x18\x06 \x01(\x02H\x00\x12\x10\n\x06\x64ouble\x18\x07 \x01(\x01H\x00\x12\x11\n\x07\x64\x65\x63imal\x18\x08 \x01(\tH\x00\x12\x11\n\x07\x62oolean\x18\t \x01(\x08H\x00\x12\x0e\n\x04\x63har\x18\n \x01(\tH\x00\x12\x11\n\x07varchar\x18\x0b \x01(\tH\x00\x12\x10\n\x06string\x18\x0c \x01(\tH\x00\x12\x10\n\x06\x62inary\x18\r \x01(\x0cH\x00\x12\x0e\n\x04\x64\x61te\x18\x0e \x01(\x05H\x00\x12\x13\n\ttimestamp\x18\x0f \x01(\x03H\x00\x12\x1b\n\x11IntervalYearMonth\x18\x10 \x01(\x03H\x00\x12\x34\n\x0fIntervalDayTime\x18\x11 \x01(\x0b\x32\x19.cz.proto.IntervalDayTimeH\x00\x12%\n\x05\x61rray\x18\x64 \x01(\x0b\x32\x14.cz.proto.ArrayValueH\x00\x12!\n\x03map\x18\x65 \x01(\x0b\x32\x12.cz.proto.MapValueH\x00\x12\'\n\x06struct\x18\x66 \x01(\x0b\x32\x15.cz.proto.StructValueH\x00\x42\x07\n\x05value\"m\n\tReference\x12\n\n\x02id\x18\x01 \x01(\x04\x12\r\n\x05local\x18\x02 \x01(\x08\x12\x0c\n\x04\x66rom\x18\x03 \x01(\t\x12\x0c\n\x04name\x18\x04 \x01(\t\x12)\n\x08ref_type\x18\x05 \x01(\x0e\x32\x17.cz.proto.ReferenceType\"\xda\x01\n\x0eScalarFunction\x12\x0c\n\x04\x66rom\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0f\n\x07\x62uiltIn\x18\x03 \x01(\x08\x12-\n\targuments\x18\x04 \x03(\x0b\x32\x1a.cz.proto.ScalarExpression\x12(\n\nproperties\x18\x05 \x01(\x0b\x32\x14.cz.proto.Properties\x12\x10\n\x08\x65xecDesc\x18\x06 \x01(\t\x12\x30\n\x12\x66unctionProperties\x18\x07 \x01(\x0b\x32\x14.cz.proto.Properties\";\n\x0bVariableDef\x12 \n\x04type\x18\x01 \x01(\x0b\x32\x12.cz.proto.DataType\x12\n\n\x02id\x18\x02 \x01(\x04\"a\n\x0eLambdaFunction\x12%\n\x06params\x18\x01 \x03(\x0b\x32\x15.cz.proto.VariableDef\x12(\n\x04impl\x18\x02 \x01(\x0b\x32\x1a.cz.proto.ScalarExpression\"\xe9\x01\n\x10ScalarExpression\x12 \n\x04type\x18\x01 \x01(\x0b\x32\x12.cz.proto.DataType\x12&\n\x08\x63onstant\x18\x02 \x01(\x0b\x32\x12.cz.proto.ConstantH\x00\x12(\n\treference\x18\x03 \x01(\x0b\x32\x13.cz.proto.ReferenceH\x00\x12,\n\x08\x66unction\x18\x05 \x01(\x0b\x32\x18.cz.proto.ScalarFunctionH\x00\x12*\n\x06lambda\x18\x06 \x01(\x0b\x32\x18.cz.proto.LambdaFunctionH\x00\x42\x07\n\x05value*W\n\rReferenceType\x12\x11\n\rLOGICAL_FIELD\x10\x00\x12\r\n\tREF_LOCAL\x10\x01\x12\x12\n\x0ePHYSICAL_FIELD\x10\x02\x12\x10\n\x0cREF_VARIABLE\x10\x03\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'expression_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_REFERENCETYPE']._serialized_start=1516
  _globals['_REFERENCETYPE']._serialized_end=1603
  _globals['_INTERVALDAYTIME']._serialized_start=63
  _globals['_INTERVALDAYTIME']._serialized_end=112
  _globals['_ARRAYVALUE']._serialized_start=114
  _globals['_ARRAYVALUE']._serialized_end=164
  _globals['_MAPVALUE']._serialized_start=166
  _globals['_MAPVALUE']._serialized_end=246
  _globals['_STRUCTVALUE']._serialized_start=248
  _globals['_STRUCTVALUE']._serialized_end=297
  _globals['_CONSTANT']._serialized_start=300
  _globals['_CONSTANT']._serialized_end=786
  _globals['_REFERENCE']._serialized_start=788
  _globals['_REFERENCE']._serialized_end=897
  _globals['_SCALARFUNCTION']._serialized_start=900
  _globals['_SCALARFUNCTION']._serialized_end=1118
  _globals['_VARIABLEDEF']._serialized_start=1120
  _globals['_VARIABLEDEF']._serialized_end=1179
  _globals['_LAMBDAFUNCTION']._serialized_start=1181
  _globals['_LAMBDAFUNCTION']._serialized_end=1278
  _globals['_SCALAREXPRESSION']._serialized_start=1281
  _globals['_SCALAREXPRESSION']._serialized_end=1514
# @@protoc_insertion_point(module_scope)



import os.path

import io
import string
import base64
from time import sleep

import oss2
import pyarrow as pa
import logging
import enum
from clickzetta.enums import FetchMode
from concurrent.futures import ThreadPoolExecutor, wait, ALL_COMPLETED, FIRST_COMPLETED

from qcloud_cos import CosConfig, CosS3Client

from clickzetta.object_storage_client import ObjectStorageClient, ObjectStorageType

logger = logging.getLogger(__name__)


class QueryDataType(enum.Enum):
    Memory = 0
    File = 1


class QueryData(object):
    def __init__(self, data: list, data_type: QueryDataType, file_list: list = None,
                 object_storage_client:ObjectStorageClient = None):
        self.data = data
        self.data_type = data_type
        self.file_list = file_list
        self.object_storage_client = object_storage_client
        self.current_file_index = 0
        self.memory_read = False

    def read(self, fetch_mode=FetchMode.FETCH_ALL, size=0):
        if self.data_type == QueryDataType.Memory:
            if self.memory_read:
                return None
            self.memory_read = True
            return self.data
        elif self.data_type == QueryDataType.File:
            assert self.file_list is not None
            try:
                final_result = []
                if fetch_mode == FetchMode.FETCH_ONE:
                    file = self.file_list[0]
                    return self.read_object_file(file)
                elif fetch_mode == FetchMode.FETCH_MANY:
                    for file in self.file_list:
                        final_result.extend(self.read_object_file(file))
                        if len(final_result) >= size:
                            break
                else:
                    for file in self.file_list:
                        final_result.extend(self.read_object_file(file))
                return final_result
            except Exception as e:
                logger.error(f'Error while converting from arrow to result: {e}')
                raise Exception(f'Error while converting from arrow to result: {e}')

    def read_object_file(self, file):
        file_info = file.split('/', 3)
        result = []
        stream = self.object_storage_client.get_stream(file_info[2], file_info[3])
        with pa.ipc.RecordBatchStreamReader(stream) as reader:
            for index, row in reader.read_pandas().iterrows():
                result.append(tuple(row.to_list()))
            return result


class Field(object):
    def __init__(self):
        self.name = None
        self.field_type = None
        self.precision = None
        self.scale = None
        self.length = None
        self.nullable = None

    def set_name(self, name):
        self.name = name

    def set_type(self, type):
        self.field_type = type

    def set_precision(self, precision):
        self.precision = precision

    def set_scale(self, scale):
        self.scale = scale

    def set_length(self, length):
        self.length = length

    def set_nullable(self, nullable):
        self.nullable = nullable


class QueryResult(object):
    def __init__(self, total_msg):
        self.data = None
        self.state = None
        self.total_row_count = 0
        self.total_msg = total_msg
        self.schema = []
        self._parse_result_data()

    def _parse_field(self, field: str, schema_field: Field):
        schema_field.set_name(field['name'])
        if field['type'].__contains__('charTypeInfo'):
            schema_field.set_type(field['type']['category'])
            schema_field.set_nullable(str(field['type']['nullable']) != 'False')
            schema_field.set_length(field['type']['charTypeInfo']['length'])
        elif field['type'].__contains__('decimalTypeInfo'):
            schema_field.set_type(field['type']['category'])
            schema_field.set_nullable(str(field['type']['nullable']) == 'true')
            schema_field.set_precision(field['type']['decimalTypeInfo']['precision'])
            schema_field.set_scale(field['type']['decimalTypeInfo']['scale'])
        else:
            schema_field.set_type(field['type']['category'])
            schema_field.set_nullable(str(field['type']['nullable']) == 'true')

    def get_result_state(self) -> string:
        return self.total_msg['status']['state']

    def get_arrow_result(self, arrow_buffer):
        try:
            buffer = base64.b64decode(arrow_buffer)
            with pa.ipc.RecordBatchStreamReader(io.BytesIO(buffer)) as reader:
                table = reader.read_all()
                column_dict = {}
                for index, column_name in enumerate(table.column_names):
                    if column_name in column_dict:
                        column_dict[f'{column_name}_{index}'] = index
                    else:
                        column_dict[column_name] = index
                new_table = table.rename_columns(column_dict.keys())
                pandas_result = new_table.to_pandas()
                self.panda_data = pandas_result
                result = []
                for index, row in pandas_result.iterrows():
                    result.append(tuple(row.tolist()))
                ##########################################################
                # result = []
                # batches = [b for b in new_table.to_batches()]
                # for batch in batches:
                #     cols = [batch.column(i) for i in range(batch.num_columns)]
                #     for row_index in range(batch.num_rows):
                #         row_values = []
                #         for col in cols:
                #             value = col[row_index].as_py()
                #             if value is not None and isinstance(col[row_index], pa.lib.MapScalar):
                #                 d = dict()
                #                 for v in value:
                #                     d[str(v[0])] = v[1]
                #                 value = d
                #             row_values.append(value)
                #         result.append(tuple(row_values))
                ############################################################
                return result

        except Exception as e:
            logger.error(f'Error while converting from arrow to result: {e}')
            raise Exception(f'Error while converting from arrow to result: {e}')

    def get_result_schema(self):
        fields = self.total_msg['resultSet']['metadata']['fields']
        for field in fields:
            schema_field = Field()
            self._parse_field(field, schema_field)
            self.schema.append(schema_field)

    def get_object_storage_type(self, path: str) -> ObjectStorageType:
        if path.lower().startswith('oss://'):
            return ObjectStorageType.OSS
        elif path.lower().startswith('cos://'):
            return ObjectStorageType.COS
        elif path.lower().startswith('s3://'):
            return ObjectStorageType.S3
        elif path.lower().startswith('gs://'):
            return ObjectStorageType.GCS
        elif path.lower().startswith("tos://"):
            return ObjectStorageType.TOS

    def get_object_storage_file_list(self) -> list:
        object_storage_file_list = self.total_msg['resultSet']['location']['location']
        return object_storage_file_list

    def get_oss_bucket(self) -> oss2.Bucket:
        location_info = self.total_msg['resultSet']['location']
        id = location_info['stsAkId']
        secret = location_info['stsAkSecret']
        token = location_info['stsToken']
        endpoint = location_info['ossEndpoint']
        if len(location_info['location']) == 0:
            raise Exception('No file found in oss when get result from clickzetta')
        bucket_name = location_info['location'][0].split('/', 3)[2]
        auth = oss2.StsAuth(id, secret, token)
        bucket = oss2.Bucket(auth, endpoint, bucket_name)
        return bucket

    def get_cos_client(self) -> CosS3Client:
        location_info = self.total_msg['resultSet']['location']
        region = location_info['objectStorageRegion']
        id = location_info['stsAkId']
        secret = location_info['stsAkSecret']
        token = location_info['stsToken']
        cos_config = CosConfig(Region=region, SecretId=id, SecretKey=secret, Token=token)
        client = CosS3Client(cos_config)
        return client

    def _parse_result_data(self):
        if len(self.total_msg) == 0:
            return
        self.state = self.total_msg['status']['state']
        if self.state != 'FAILED' and self.state != 'CANCELLED':
            if 'data' not in self.total_msg['resultSet']:
                if 'location' in self.total_msg['resultSet']:
                    self.get_result_schema()
                    file_list = self.get_object_storage_file_list()
                    if len(file_list) == 0:
                        self.total_row_count = 0
                        self.data = QueryData(data=[], data_type=QueryDataType.Memory)
                        return
                    object_storage_type = self.get_object_storage_type(file_list[0])
                    location_info = self.total_msg['resultSet']['location']
                    ak_id = location_info['stsAkId']
                    ak_secret = location_info['stsAkSecret']
                    token = location_info['stsToken']
                    endpoint = None
                    bucket = None
                    if object_storage_type == ObjectStorageType.OSS:
                        endpoint = location_info['ossEndpoint']
                        bucket = location_info['location'][0].split('/', 3)[2]
                    elif object_storage_type == ObjectStorageType.COS:
                        endpoint = location_info['objectStorageRegion']
                    elif object_storage_type == ObjectStorageType.S3:
                        endpoint = location_info['objectStorageRegion']
                    else:
                        endpoint = location_info["objectStorageRegion"]
                    object_storage_client = ObjectStorageClient(object_storage_type, ak_id, ak_secret, token,
                                                                endpoint, bucket)
                    self.data = QueryData(data_type=QueryDataType.File, file_list=file_list, data=None,
                                          object_storage_client=object_storage_client)

                else:
                    field = Field()
                    field.set_name('RESULT_MESSAGE')
                    field.set_type("STRING")
                    self.schema.append(field)
                    self.total_row_count = 1
                    result_data = [['OPERATION SUCCEED']]
                    self.data = QueryData(data=result_data, data_type=QueryDataType.Memory)
            else:
                if not (len(self.total_msg['resultSet']['data']['data'])):
                    self.total_row_count = 0
                    fields = self.total_msg['resultSet']['metadata']['fields']
                    for field in fields:
                        schema_field = Field()
                        self._parse_field(field, schema_field)
                        self.schema.append(schema_field)
                    self.data = QueryData(data=[], data_type=QueryDataType.Memory)
                    return
                result_data = self.total_msg['resultSet']['data']['data']
                self.get_result_schema()
                query_result = []
                for row in result_data:
                    partial_result = self.get_arrow_result(row)
                    query_result.extend(entity for entity in partial_result)
                self.data = QueryData(data=query_result, data_type=QueryDataType.Memory)

        else:
            raise Exception('SQL job execute failed.Error:' + self.total_msg['status']['message'])



"""Connection for ClickZetta DB-API."""
import weakref
import logging
import time

import requests
from requests.adapters import HTTPAdapter
from clickzetta.client import Client
from clickzetta.dbapi import cursor
from clickzetta.dbapi import _helpers
from clickzetta.bulkload.bulkload_enums import BulkLoadOperation, BulkLoadOptions, BulkLoadCommitOptions
from clickzetta.bulkload.bulkload_stream import BulkLoadStream

is_token_init = False
current_token = None
token_init_time = 0
https_session = None
user_token_dict = {}
user_token_time_dict = {}
https_session_inited = False


@_helpers.raise_on_closed("Operating on a closed connection.")
class Connection(object):
    def __init__(self, client=None):
        if client is None:
            logging.error('Connection must has a LogParams to log in.')
            raise AssertionError('Connection must has a LogParams to log in.')
        else:
            self._owns_client = True

        self._client = client
        if self._client.magic_token is not None:
            self._client.token = self._client.magic_token
        else:
            self._update_token()

        if not globals()['https_session_inited']:
            session = requests.Session()
            session.mount(self._client.service, HTTPAdapter(pool_connections=20, pool_maxsize=100, max_retries=1000))
            globals()['https_session'] = session
            globals()['https_session_inited'] = True

        self._client.session = globals()['https_session']
        self._closed = False
        self._cursors_created = weakref.WeakSet()

    def _update_token(self):
        current_time = int(time.time())
        unique_key = self._client.username + "?" + self._client.instance
        if globals()['user_token_dict'].__contains__(unique_key):
            if current_time - globals()['user_token_time_dict'][unique_key] > 24 * 60 * 60 * 2:
                new_token = self._client.log_in_cz(self._client.username, self._client.password,
                                                   self._client.instance)
                globals()['user_token_dict'][unique_key] = new_token
                globals()['user_token_time_dict'][unique_key] = current_time
                self._client.token = new_token
            else:
                self._client.token = globals()['user_token_dict'][unique_key]
        else:
            new_token = self._client.log_in_cz(self._client.username, self._client.password, self._client.instance)
            globals()['user_token_dict'][unique_key] = new_token
            globals()['user_token_time_dict'][unique_key] = current_time
            self._client.token = new_token

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        self._closed = True

        if self._owns_client:
            self._client.close()

        for cursor_ in self._cursors_created:
            if not cursor_._closed:
                cursor_.close()

    def commit(self):
        """No-op, but for consistency raise an error if connection is closed."""

    def cursor(self):
        """Return a new cursor object."""
        if self._client.username is not None and self._client.password is not None:
            self._update_token()
        new_cursor = cursor.Cursor(self)
        self._cursors_created.add(new_cursor)
        return new_cursor

    def create_bulkload_stream(self, **kwargs):
        schema = kwargs.get('schema', self._client.schema)
        table = kwargs.get('table')
        if schema is None:
            schema = self._client.schema
        if schema is None:
            raise ValueError(f'No schema specified')
        if table is None:
            raise ValueError(f'No table specified')

        operation = kwargs.get('operation', BulkLoadOperation.APPEND)
        workspace = kwargs.get('workspace', self._client.workspace)
        vcluster = kwargs.get('vcluster', self._client.vcluster)
        partition_spec = kwargs.get('partition_spec')
        record_keys = kwargs.get('record_keys')
        prefer_internal_endpoint = kwargs.get('prefer_internal_endpoint', False)

        bulkload_meta_data = self._client.create_bulkload_stream(
            schema, table, BulkLoadOptions(operation, partition_spec, record_keys, prefer_internal_endpoint))
        return BulkLoadStream(bulkload_meta_data, self._client,
                              BulkLoadCommitOptions(workspace, vcluster))

    def get_bulkload_stream(self, stream_id: str, schema: str = None, table: str = None):
        bulkload_meta_data = self._client.get_bulkload_stream(schema, table, stream_id)
        return BulkLoadStream(bulkload_meta_data, self._client,
                              BulkLoadCommitOptions(self._client.workspace, self._client.vcluster))

    def get_job_profile(self, job_id: str):
        return self._client.get_job_profile(job_id)

    def get_job_result(self, job_id: str):
        return self._client.get_job_result(job_id)

    def get_job_progress(self, job_id: str):
        return self._client.get_job_progress(job_id)

    def get_job_summary(self, job_id: str):
        return self._client.get_job_summary(job_id)

    def get_job_plan(self, job_id: str):
        return self._client.get_job_plan(job_id)


def connect(**kwargs) -> Connection:
    client = kwargs.get('client')
    if client is None:
        client = Client(cz_url=kwargs.get('cz_url'),  # setting client or cz_url will ignore following parameters
                        username=kwargs.get('username'), password=kwargs.get('password'),
                        instance=kwargs.get('instance'), service=kwargs.get('service'),
                        workspace=kwargs.get('workspace'), vcluster=kwargs.get('vcluster'),
                        schema=kwargs.get('schema', 'public'), magic_token=kwargs.get('magic_token'),
                        protocol=kwargs.get('protocol'), hints=kwargs.get('hints'))
    return Connection(client)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: hash.proto
# Protobuf Python Version: 5.28.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    1,
    '',
    'hash.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\nhash.proto\x12\x04kudu*R\n\rHashAlgorithm\x12\x10\n\x0cUNKNOWN_HASH\x10\x00\x12\x11\n\rMURMUR_HASH_2\x10\x01\x12\r\n\tCITY_HASH\x10\x02\x12\r\n\tFAST_HASH\x10\x03\x42\x11\n\x0forg.apache.kudu')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'hash_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\017org.apache.kudu'
  _globals['_HASHALGORITHM']._serialized_start=20
  _globals['_HASHALGORITHM']._serialized_end=102
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: table_common.proto
# Protobuf Python Version: 5.28.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    1,
    '',
    'table_common.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import data_type_pb2 as data__type__pb2
from . import file_format_type_pb2 as file__format__type__pb2
from . import file_system_pb2 as file__system__pb2
from . import object_identifier_pb2 as object__identifier__pb2
from . import expression_pb2 as expression__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x12table_common.proto\x12\x08\x63z.proto\x1a\x0f\x64\x61ta_type.proto\x1a\x16\x66ile_format_type.proto\x1a\x11\x66ile_system.proto\x1a\x17object_identifier.proto\x1a\x10\x65xpression.proto\"=\n\x08\x46ieldRef\x12\x12\n\x08\x66ield_id\x18\x01 \x01(\rH\x00\x12\x14\n\nfield_name\x18\x02 \x01(\tH\x00\x42\x07\n\x05\x66ield\"U\n\x0bSortedField\x12!\n\x05\x66ield\x18\x01 \x01(\x0b\x32\x12.cz.proto.FieldRef\x12#\n\nsort_order\x18\x02 \x01(\x0e\x32\x0f.cz.proto.Order\"V\n\x0bHashCluster\x12\x18\n\x10\x66unction_version\x18\x01 \x01(\r\x12-\n\x0b\x62ucket_type\x18\x02 \x01(\x0e\x32\x18.cz.proto.HashBucketType\"7\n\x0cRangeCluster\x12\'\n\nrange_type\x18\x01 \x01(\x0e\x32\x13.cz.proto.RangeType\"\x9d\x02\n\x0b\x43lusterInfo\x12+\n\x0c\x63luster_type\x18\x01 \x01(\x0e\x32\x15.cz.proto.ClusterType\x12,\n\x10\x63lustered_fields\x18\x02 \x03(\x0b\x32\x12.cz.proto.FieldRef\x12\x1a\n\rbuckets_count\x18\x03 \x01(\x04H\x01\x88\x01\x01\x12\x19\n\x0cpath_pattern\x18\x04 \x01(\tH\x02\x88\x01\x01\x12%\n\x04hash\x18\n \x01(\x0b\x32\x15.cz.proto.HashClusterH\x00\x12\'\n\x05range\x18\x0b \x01(\x0b\x32\x16.cz.proto.RangeClusterH\x00\x42\t\n\x07\x63lusterB\x10\n\x0e_buckets_countB\x0f\n\r_path_pattern\"9\n\tSortOrder\x12,\n\rsorted_fields\x18\x01 \x03(\x0b\x32\x15.cz.proto.SortedField\"f\n\tUniqueKey\x12)\n\runique_fields\x18\x01 \x03(\x0b\x32\x12.cz.proto.FieldRef\x12\x0e\n\x06\x65nable\x18\x02 \x01(\x08\x12\x10\n\x08validate\x18\x03 \x01(\x08\x12\x0c\n\x04rely\x18\x04 \x01(\x08\"`\n\nPrimaryKey\x12\"\n\x06\x66ields\x18\x01 \x03(\x0b\x32\x12.cz.proto.FieldRef\x12\x0e\n\x06\x65nable\x18\x02 \x01(\x08\x12\x10\n\x08validate\x18\x03 \x01(\x08\x12\x0c\n\x04rely\x18\x04 \x01(\x08\".\n\x08IndexKey\x12\"\n\x06\x66ields\x18\x01 \x03(\x0b\x32\x12.cz.proto.FieldRef\"v\n\x05Index\x12!\n\x04type\x18\x01 \x01(\x0e\x32\x13.cz.proto.IndexType\x12\x1f\n\x03key\x18\x02 \x01(\x0b\x32\x12.cz.proto.IndexKey\x12)\n\x05table\x18\x03 \x01(\x0b\x32\x1a.cz.proto.ObjectIdentifier\"\xf8\x01\n\tFieldSpec\x12\x0f\n\x07spec_id\x18\x01 \x01(\r\x12-\n\x0c\x63luster_info\x18\n \x01(\x0b\x32\x15.cz.proto.ClusterInfoH\x00\x12)\n\nsort_order\x18\x0b \x01(\x0b\x32\x13.cz.proto.SortOrderH\x00\x12)\n\nunique_key\x18\x0c \x01(\x0b\x32\x13.cz.proto.UniqueKeyH\x00\x12+\n\x0bprimary_key\x18\r \x01(\x0b\x32\x14.cz.proto.PrimaryKeyH\x00\x12 \n\x05index\x18\x0e \x01(\x0b\x32\x0f.cz.proto.IndexH\x00\x42\x06\n\x04spec\"\x86\x02\n\x0b\x46ieldSchema\x12\x0c\n\x04name\x18\x02 \x01(\t\x12 \n\x04type\x18\x03 \x01(\x0b\x32\x12.cz.proto.DataType\x12\x14\n\x07virtual\x18\x04 \x01(\x08H\x00\x88\x01\x01\x12\x13\n\x06hidden\x18\x05 \x01(\x08H\x01\x88\x01\x01\x12\x16\n\tun_output\x18\x06 \x01(\x08H\x02\x88\x01\x01\x12\x0f\n\x07\x63omment\x18\x07 \x01(\t\x12(\n\x04\x65xpr\x18\x08 \x01(\x0b\x32\x1a.cz.proto.ScalarExpression\x12\x16\n\ttransform\x18\t \x01(\tH\x03\x88\x01\x01\x42\n\n\x08_virtualB\t\n\x07_hiddenB\x0c\n\n_un_outputB\x0c\n\n_transform\"U\n\x0bTableSchema\x12%\n\x06\x66ields\x18\x01 \x03(\x0b\x32\x15.cz.proto.FieldSchema\x12\x11\n\tschema_id\x18\x02 \x01(\r\x12\x0c\n\x04type\x18\x03 \x01(\t\"x\n\x0eTextFileFormat\x12\x36\n\x07options\x18\x01 \x03(\x0b\x32%.cz.proto.TextFileFormat.OptionsEntry\x1a.\n\x0cOptionsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"c\n\x11ParquetFileFormat\x12\x1c\n\x14row_group_size_bytes\x18\x01 \x01(\x03\x12\x17\n\x0fpage_size_bytes\x18\x02 \x01(\x03\x12\x17\n\x0f\x64ict_size_bytes\x18\x03 \x01(\x03\"\x0f\n\rOrcFileFormat\"v\n\rCsvFileFormat\x12\x35\n\x07options\x18\x01 \x03(\x0b\x32$.cz.proto.CsvFileFormat.OptionsEntry\x1a.\n\x0cOptionsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\x84\x01\n\x14HiveResultFileFormat\x12<\n\x07options\x18\x01 \x03(\x0b\x32+.cz.proto.HiveResultFileFormat.OptionsEntry\x1a.\n\x0cOptionsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\x10\n\x0e\x41vroFileFormat\"\x11\n\x0f\x41rrowFileFormat\"\x97\x03\n\nFileFormat\x12&\n\x04type\x18\x01 \x01(\x0e\x32\x18.cz.proto.FileFormatType\x12,\n\x08textFile\x18\x02 \x01(\x0b\x32\x18.cz.proto.TextFileFormatH\x00\x12\x33\n\x0cparquet_file\x18\x03 \x01(\x0b\x32\x1b.cz.proto.ParquetFileFormatH\x00\x12+\n\x08orc_file\x18\x04 \x01(\x0b\x32\x17.cz.proto.OrcFileFormatH\x00\x12+\n\x08\x63sv_file\x18\x05 \x01(\x0b\x32\x17.cz.proto.CsvFileFormatH\x00\x12:\n\x10hive_result_file\x18\x06 \x01(\x0b\x32\x1e.cz.proto.HiveResultFileFormatH\x00\x12-\n\tavro_file\x18\x07 \x01(\x0b\x32\x18.cz.proto.AvroFileFormatH\x00\x12/\n\narrow_file\x18\x08 \x01(\x0b\x32\x19.cz.proto.ArrowFileFormatH\x00\x42\x08\n\x06\x66ormat\"z\n\x12\x46ileDataSourceInfo\x12\x30\n\x0e\x66ileSystemType\x18\x01 \x01(\x0e\x32\x18.cz.proto.FileSystemType\x12\x0c\n\x04path\x18\x02 \x01(\t\x12$\n\x06\x66ormat\x18\x03 \x01(\x0b\x32\x14.cz.proto.FileFormat\"\xca\x01\n\x0e\x44\x61taSourceInfo\x12,\n\x04\x66ile\x18\x01 \x01(\x0b\x32\x1c.cz.proto.FileDataSourceInfoH\x00\x12\x36\n\x07options\x18\x02 \x03(\x0b\x32%.cz.proto.DataSourceInfo.OptionsEntry\x12\x10\n\x08location\x18\x03 \x01(\t\x1a.\n\x0cOptionsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x42\x10\n\x0e\x64\x61taSourceInfo\"a\n\nDataSource\x12\x33\n\x11\x64\x61ta_source_infos\x18\x11 \x03(\x0b\x32\x18.cz.proto.DataSourceInfo\x12\x1e\n\x16\x64\x65\x66\x61ult_data_source_id\x18\x12 \x01(\r\"R\n\x08MVSource\x12\x34\n\x10table_identifier\x18\x01 \x01(\x0b\x32\x1a.cz.proto.ObjectIdentifier\x12\x10\n\x08snapshot\x18\x02 \x01(\x03\"\xa2\x01\n\rRefreshOption\x12*\n\x04type\x18\x01 \x01(\x0e\x32\x1c.cz.proto.RefreshOption.Type\x12\x12\n\nstart_time\x18\x02 \x01(\x03\x12\x1a\n\x12interval_in_minute\x18\x03 \x01(\x03\"5\n\x04Type\x12\r\n\tON_DEMAND\x10\x00\x12\r\n\tON_COMMIT\x10\x01\x12\x0f\n\x0bON_SCHEDULE\x10\x02\"\xad\x01\n\x0bMVExtension\x12\x0f\n\x07mv_plan\x18\x01 \x01(\t\x12,\n\x10mv_source_tables\x18\x02 \x03(\x0b\x32\x12.cz.proto.MVSource\x12/\n\x0erefresh_option\x18\x03 \x01(\x0b\x32\x17.cz.proto.RefreshOption\x12\x1b\n\x0emv_snapshot_id\x18\x04 \x01(\x03H\x00\x88\x01\x01\x42\x11\n\x0f_mv_snapshot_id\"k\n\x04View\x12\x1a\n\x12view_expanded_text\x18\x01 \x01(\t\x12\x1a\n\x12view_original_text\x18\x02 \x01(\t\x12+\n\x0cmv_extension\x18\x05 \x01(\x0b\x32\x15.cz.proto.MVExtension*\x1a\n\x05Order\x12\x07\n\x03\x41SC\x10\x00\x12\x08\n\x04\x44\x45SC\x10\x01*.\n\x0b\x43lusterType\x12\n\n\x06NORMAL\x10\x00\x12\t\n\x05RANGE\x10\x01\x12\x08\n\x04HASH\x10\x02*.\n\x0eHashBucketType\x12\x0c\n\x08HASH_MOD\x10\x00\x12\x0e\n\nHASH_RANGE\x10\x01*n\n\tTableType\x12\x11\n\rMANAGED_TABLE\x10\x00\x12\x12\n\x0e\x45XTERNAL_TABLE\x10\x02\x12\x10\n\x0cVIRTUAL_VIEW\x10\x04\x12\x15\n\x11MATERIALIZED_VIEW\x10\x06\x12\x11\n\rUNKNOWN_TABLE\x10\x63*\x1d\n\tIndexType\x12\x10\n\x0c\x42LOOM_FILTER\x10\x00*@\n\tRangeType\x12\x0f\n\x0b\x46IXED_POINT\x10\x00\x12\x0f\n\x0b\x46IXED_RANGE\x10\x02\x12\x11\n\rDYNAMIC_RANGE\x10\x04\x42\x0c\n\x08\x63z.protoP\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'table_common_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\010cz.protoP\001'
  _globals['_TEXTFILEFORMAT_OPTIONSENTRY']._loaded_options = None
  _globals['_TEXTFILEFORMAT_OPTIONSENTRY']._serialized_options = b'8\001'
  _globals['_CSVFILEFORMAT_OPTIONSENTRY']._loaded_options = None
  _globals['_CSVFILEFORMAT_OPTIONSENTRY']._serialized_options = b'8\001'
  _globals['_HIVERESULTFILEFORMAT_OPTIONSENTRY']._loaded_options = None
  _globals['_HIVERESULTFILEFORMAT_OPTIONSENTRY']._serialized_options = b'8\001'
  _globals['_DATASOURCEINFO_OPTIONSENTRY']._loaded_options = None
  _globals['_DATASOURCEINFO_OPTIONSENTRY']._serialized_options = b'8\001'
  _globals['_ORDER']._serialized_start=3654
  _globals['_ORDER']._serialized_end=3680
  _globals['_CLUSTERTYPE']._serialized_start=3682
  _globals['_CLUSTERTYPE']._serialized_end=3728
  _globals['_HASHBUCKETTYPE']._serialized_start=3730
  _globals['_HASHBUCKETTYPE']._serialized_end=3776
  _globals['_TABLETYPE']._serialized_start=3778
  _globals['_TABLETYPE']._serialized_end=3888
  _globals['_INDEXTYPE']._serialized_start=3890
  _globals['_INDEXTYPE']._serialized_end=3919
  _globals['_RANGETYPE']._serialized_start=3921
  _globals['_RANGETYPE']._serialized_end=3985
  _globals['_FIELDREF']._serialized_start=135
  _globals['_FIELDREF']._serialized_end=196
  _globals['_SORTEDFIELD']._serialized_start=198
  _globals['_SORTEDFIELD']._serialized_end=283
  _globals['_HASHCLUSTER']._serialized_start=285
  _globals['_HASHCLUSTER']._serialized_end=371
  _globals['_RANGECLUSTER']._serialized_start=373
  _globals['_RANGECLUSTER']._serialized_end=428
  _globals['_CLUSTERINFO']._serialized_start=431
  _globals['_CLUSTERINFO']._serialized_end=716
  _globals['_SORTORDER']._serialized_start=718
  _globals['_SORTORDER']._serialized_end=775
  _globals['_UNIQUEKEY']._serialized_start=777
  _globals['_UNIQUEKEY']._serialized_end=879
  _globals['_PRIMARYKEY']._serialized_start=881
  _globals['_PRIMARYKEY']._serialized_end=977
  _globals['_INDEXKEY']._serialized_start=979
  _globals['_INDEXKEY']._serialized_end=1025
  _globals['_INDEX']._serialized_start=1027
  _globals['_INDEX']._serialized_end=1145
  _globals['_FIELDSPEC']._serialized_start=1148
  _globals['_FIELDSPEC']._serialized_end=1396
  _globals['_FIELDSCHEMA']._serialized_start=1399
  _globals['_FIELDSCHEMA']._serialized_end=1661
  _globals['_TABLESCHEMA']._serialized_start=1663
  _globals['_TABLESCHEMA']._serialized_end=1748
  _globals['_TEXTFILEFORMAT']._serialized_start=1750
  _globals['_TEXTFILEFORMAT']._serialized_end=1870
  _globals['_TEXTFILEFORMAT_OPTIONSENTRY']._serialized_start=1824
  _globals['_TEXTFILEFORMAT_OPTIONSENTRY']._serialized_end=1870
  _globals['_PARQUETFILEFORMAT']._serialized_start=1872
  _globals['_PARQUETFILEFORMAT']._serialized_end=1971
  _globals['_ORCFILEFORMAT']._serialized_start=1973
  _globals['_ORCFILEFORMAT']._serialized_end=1988
  _globals['_CSVFILEFORMAT']._serialized_start=1990
  _globals['_CSVFILEFORMAT']._serialized_end=2108
  _globals['_CSVFILEFORMAT_OPTIONSENTRY']._serialized_start=1824
  _globals['_CSVFILEFORMAT_OPTIONSENTRY']._serialized_end=1870
  _globals['_HIVERESULTFILEFORMAT']._serialized_start=2111
  _globals['_HIVERESULTFILEFORMAT']._serialized_end=2243
  _globals['_HIVERESULTFILEFORMAT_OPTIONSENTRY']._serialized_start=1824
  _globals['_HIVERESULTFILEFORMAT_OPTIONSENTRY']._serialized_end=1870
  _globals['_AVROFILEFORMAT']._serialized_start=2245
  _globals['_AVROFILEFORMAT']._serialized_end=2261
  _globals['_ARROWFILEFORMAT']._serialized_start=2263
  _globals['_ARROWFILEFORMAT']._serialized_end=2280
  _globals['_FILEFORMAT']._serialized_start=2283
  _globals['_FILEFORMAT']._serialized_end=2690
  _globals['_FILEDATASOURCEINFO']._serialized_start=2692
  _globals['_FILEDATASOURCEINFO']._serialized_end=2814
  _globals['_DATASOURCEINFO']._serialized_start=2817
  _globals['_DATASOURCEINFO']._serialized_end=3019
  _globals['_DATASOURCEINFO_OPTIONSENTRY']._serialized_start=1824
  _globals['_DATASOURCEINFO_OPTIONSENTRY']._serialized_end=1870
  _globals['_DATASOURCE']._serialized_start=3021
  _globals['_DATASOURCE']._serialized_end=3118
  _globals['_MVSOURCE']._serialized_start=3120
  _globals['_MVSOURCE']._serialized_end=3202
  _globals['_REFRESHOPTION']._serialized_start=3205
  _globals['_REFRESHOPTION']._serialized_end=3367
  _globals['_REFRESHOPTION_TYPE']._serialized_start=3314
  _globals['_REFRESHOPTION_TYPE']._serialized_end=3367
  _globals['_MVEXTENSION']._serialized_start=3370
  _globals['_MVEXTENSION']._serialized_end=3543
  _globals['_VIEW']._serialized_start=3545
  _globals['_VIEW']._serialized_end=3652
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: account.proto
# Protobuf Python Version: 5.28.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    1,
    '',
    'account.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import object_identifier_pb2 as object__identifier__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\raccount.proto\x12\x08\x63z.proto\x1a\x17object_identifier.proto\"V\n\x0fPolicyStatement\x12\x0e\n\x06\x45\x66\x66\x65\x63t\x18\x01 \x01(\t\x12\x11\n\tPrincipal\x18\x02 \x03(\t\x12\x0e\n\x06\x41\x63tion\x18\x03 \x03(\t\x12\x10\n\x08Resource\x18\x04 \x03(\t\"G\n\x06Policy\x12\x0f\n\x07Version\x18\x01 \x01(\t\x12,\n\tStatement\x18\x02 \x03(\x0b\x32\x19.cz.proto.PolicyStatement\"\xa8\x01\n\x07\x41\x63\x63ount\x12\x12\n\naccount_id\x18\x01 \x01(\x03\x12\x11\n\tuser_name\x18\x02 \x01(\t\x12\x0f\n\x07user_id\x18\x03 \x01(\x03\x12*\n\x04type\x18\x04 \x01(\x0e\x32\x17.cz.proto.PrincipalTypeH\x00\x88\x01\x01\x12%\n\x06policy\x18\x05 \x01(\x0b\x32\x10.cz.proto.PolicyH\x01\x88\x01\x01\x42\x07\n\x05_typeB\t\n\x07_policy\"`\n\x08Instance\x12\x14\n\x0c\x61\x63\x63ount_name\x18\x04 \x01(\t\x12\x12\n\naccount_id\x18\x01 \x01(\x03\x12\x13\n\x0binstance_id\x18\x02 \x01(\x03\x12\x15\n\rinstance_name\x18\x03 \x01(\t\"Q\n\x0eUserIdentifier\x12.\n\nidentifier\x18\x01 \x01(\x0b\x32\x1a.cz.proto.ObjectIdentifier\x12\x0f\n\x07user_id\x18\x02 \x01(\x03\"T\n\x04User\x12\x14\n\x07user_id\x18\x01 \x01(\x03H\x00\x88\x01\x01\x12\x12\n\ndefault_vc\x18\x02 \x01(\t\x12\x16\n\x0e\x64\x65\x66\x61ult_schema\x18\x03 \x01(\tB\n\n\x08_user_idb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'account_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_POLICYSTATEMENT']._serialized_start=52
  _globals['_POLICYSTATEMENT']._serialized_end=138
  _globals['_POLICY']._serialized_start=140
  _globals['_POLICY']._serialized_end=211
  _globals['_ACCOUNT']._serialized_start=214
  _globals['_ACCOUNT']._serialized_end=382
  _globals['_INSTANCE']._serialized_start=384
  _globals['_INSTANCE']._serialized_end=480
  _globals['_USERIDENTIFIER']._serialized_start=482
  _globals['_USERIDENTIFIER']._serialized_end=563
  _globals['_USER']._serialized_start=565
  _globals['_USER']._serialized_end=649
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: ingestion.proto
# Protobuf Python Version: 5.28.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    1,
    '',
    'ingestion.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import data_type_pb2 as data__type__pb2
from . import file_format_type_pb2 as file__format__type__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0fingestion.proto\x12\x12\x63z.proto.ingestion\x1a\x0f\x64\x61ta_type.proto\x1a\x16\x66ile_format_type.proto\"^\n\x0eGatewayRequest\x12\x17\n\x0fmethodEnumValue\x18\x01 \x01(\x05\x12\x0f\n\x07message\x18\x02 \x01(\t\x12\x12\n\ninstanceId\x18\x03 \x01(\x03\x12\x0e\n\x06userId\x18\x04 \x01(\x03\"]\n\x0fGatewayResponse\x12\x39\n\x06status\x18\x01 \x01(\x0b\x32).cz.proto.ingestion.GateWayResponseStatus\x12\x0f\n\x07message\x18\x02 \x01(\t\"\\\n\x0eUserIdentifier\x12\x13\n\x0binstance_id\x18\x01 \x01(\x03\x12\x11\n\tworkspace\x18\x02 \x01(\t\x12\x11\n\tuser_name\x18\x03 \x01(\t\x12\x0f\n\x07user_id\x18\x04 \x01(\x03\"P\n\x07\x41\x63\x63ount\x12\x36\n\nuser_ident\x18\x01 \x01(\x0b\x32\".cz.proto.ingestion.UserIdentifier\x12\r\n\x05token\x18\x02 \x01(\t\"b\n\x0fTableIdentifier\x12\x13\n\x0binstance_id\x18\x01 \x01(\x03\x12\x11\n\tworkspace\x18\x02 \x01(\t\x12\x13\n\x0bschema_name\x18\x03 \x01(\t\x12\x12\n\ntable_name\x18\x04 \x01(\t\"d\n\x15GateWayResponseStatus\x12&\n\x04\x63ode\x18\x01 \x01(\x0e\x32\x18.cz.proto.ingestion.Code\x12\x0f\n\x07message\x18\x02 \x01(\t\x12\x12\n\nrequest_id\x18\x03 \x01(\t\"c\n\x0eResponseStatus\x12&\n\x04\x63ode\x18\x01 \x01(\x0e\x32\x18.cz.proto.ingestion.Code\x12\x15\n\rerror_message\x18\x02 \x01(\t\x12\x12\n\nrequest_id\x18\x03 \x01(\t\";\n\tDataField\x12\x0c\n\x04name\x18\x01 \x01(\t\x12 \n\x04type\x18\x02 \x01(\x0b\x32\x12.cz.proto.DataType\"R\n\x10\x44istributionSpec\x12\x11\n\tfield_ids\x18\x01 \x03(\r\x12\x16\n\x0ehash_functions\x18\x02 \x03(\t\x12\x13\n\x0bnum_buckets\x18\x03 \x01(\r\"#\n\x0ePrimaryKeySpec\x12\x11\n\tfield_ids\x18\x01 \x03(\r\"&\n\rPartitionSpec\x12\x15\n\rsrc_field_ids\x18\x01 \x03(\r\"\xf4\x01\n\x0cStreamSchema\x12\x32\n\x0b\x64\x61ta_fields\x18\x01 \x03(\x0b\x32\x1d.cz.proto.ingestion.DataField\x12\x37\n\tdist_spec\x18\x02 \x01(\x0b\x32$.cz.proto.ingestion.DistributionSpec\x12<\n\x10primary_key_spec\x18\x03 \x01(\x0b\x32\".cz.proto.ingestion.PrimaryKeySpec\x12\x39\n\x0epartition_spec\x18\x04 \x01(\x0b\x32!.cz.proto.ingestion.PartitionSpec\"\x97\x01\n\x18\x43reateOrGetStreamRequest\x12,\n\x07\x61\x63\x63ount\x18\x01 \x01(\x0b\x32\x1b.cz.proto.ingestion.Account\x12\x38\n\x0btable_ident\x18\x02 \x01(\x0b\x32#.cz.proto.ingestion.TableIdentifier\x12\x13\n\x0bnum_tablets\x18\x03 \x01(\r\"\xd8\x01\n\x19\x43reateOrGetStreamResponse\x12\x38\n\x0btable_ident\x18\x01 \x01(\x0b\x32#.cz.proto.ingestion.TableIdentifier\x12\x35\n\x0b\x64\x61ta_schema\x18\x02 \x01(\x0b\x32 .cz.proto.ingestion.StreamSchema\x12\x16\n\x0e\x61lready_exists\x18\x03 \x01(\x08\x12\x32\n\x06status\x18\x04 \x01(\x0b\x32\".cz.proto.ingestion.ResponseStatus\"|\n\x12\x43loseStreamRequest\x12,\n\x07\x61\x63\x63ount\x18\x01 \x01(\x0b\x32\x1b.cz.proto.ingestion.Account\x12\x38\n\x0btable_ident\x18\x02 \x01(\x0b\x32#.cz.proto.ingestion.TableIdentifier\"I\n\x13\x43loseStreamResponse\x12\x32\n\x06status\x18\x01 \x01(\x0b\x32\".cz.proto.ingestion.ResponseStatus\"\xad\x01\n\x12OssStagingPathInfo\x12\x0c\n\x04path\x18\x01 \x01(\t\x12\x11\n\tsts_ak_id\x18\x03 \x01(\t\x12\x15\n\rsts_ak_secret\x18\x04 \x01(\t\x12\x11\n\tsts_token\x18\x05 \x01(\t\x12\x14\n\x0coss_endpoint\x18\x06 \x01(\t\x12\x1d\n\x15oss_internal_endpoint\x18\x07 \x01(\t\x12\x17\n\x0foss_expire_time\x18\x08 \x01(\x04\"\x8c\x01\n\x12\x43osStagingPathInfo\x12\x0c\n\x04path\x18\x01 \x01(\t\x12\x11\n\tsts_ak_id\x18\x02 \x01(\t\x12\x15\n\rsts_ak_secret\x18\x03 \x01(\t\x12\x11\n\tsts_token\x18\x04 \x01(\t\x12\x12\n\ncos_region\x18\x05 \x01(\t\x12\x17\n\x0f\x63os_expire_time\x18\x06 \x01(\x04\"\x89\x01\n\x11S3StagingPathInfo\x12\x0c\n\x04path\x18\x01 \x01(\t\x12\x11\n\tsts_ak_id\x18\x02 \x01(\t\x12\x15\n\rsts_ak_secret\x18\x03 \x01(\t\x12\x11\n\tsts_token\x18\x04 \x01(\t\x12\x11\n\ts3_region\x18\x05 \x01(\t\x12\x16\n\x0es3_expire_time\x18\x06 \x01(\x04\"\x8c\x01\n\x12GcsStagingPathInfo\x12\x0c\n\x04path\x18\x01 \x01(\t\x12\x11\n\tsts_ak_id\x18\x02 \x01(\t\x12\x15\n\rsts_ak_secret\x18\x03 \x01(\t\x12\x11\n\tsts_token\x18\x04 \x01(\t\x12\x12\n\ngcs_region\x18\x05 \x01(\t\x12\x17\n\x0fgcs_expire_time\x18\x06 \x01(\x04\"\x8e\x01\n\x12ObsStagingPathInfo\x12\x0c\n\x04path\x18\x01 \x01(\t\x12\x11\n\tsts_ak_id\x18\x02 \x01(\t\x12\x15\n\rsts_ak_secret\x18\x03 \x01(\t\x12\x11\n\tsts_token\x18\x04 \x01(\t\x12\x14\n\x0cobs_endpoint\x18\x05 \x01(\t\x12\x17\n\x0fobs_expire_time\x18\x06 \x01(\x04\"\x8e\x01\n\x12TosStagingPathInfo\x12\x0c\n\x04path\x18\x01 \x01(\t\x12\x11\n\tsts_ak_id\x18\x02 \x01(\t\x12\x15\n\rsts_ak_secret\x18\x03 \x01(\t\x12\x11\n\tsts_token\x18\x04 \x01(\t\x12\x14\n\x0ctos_endpoint\x18\x05 \x01(\t\x12\x17\n\x0ftos_expire_time\x18\x06 \x01(\x04\"$\n\x14LocalStagingPathInfo\x12\x0c\n\x04path\x18\x01 \x01(\t\"\xc4\x03\n\x0fStagingPathInfo\x12:\n\x08oss_path\x18\x01 \x01(\x0b\x32&.cz.proto.ingestion.OssStagingPathInfoH\x00\x12:\n\x08\x63os_path\x18\x02 \x01(\x0b\x32&.cz.proto.ingestion.CosStagingPathInfoH\x00\x12>\n\nlocal_path\x18\x03 \x01(\x0b\x32(.cz.proto.ingestion.LocalStagingPathInfoH\x00\x12\x38\n\x07s3_path\x18\x04 \x01(\x0b\x32%.cz.proto.ingestion.S3StagingPathInfoH\x00\x12:\n\x08gcs_path\x18\x05 \x01(\x0b\x32&.cz.proto.ingestion.GcsStagingPathInfoH\x00\x12:\n\x08obs_path\x18\x06 \x01(\x0b\x32&.cz.proto.ingestion.ObsStagingPathInfoH\x00\x12:\n\x08tos_path\x18\x07 \x01(\x0b\x32&.cz.proto.ingestion.TosStagingPathInfoH\x00\x42\x0b\n\tpath_info\"\xa7\x04\n\x12\x42ulkLoadStreamInfo\x12\x11\n\tstream_id\x18\x01 \x01(\t\x12=\n\x0cstream_state\x18\x02 \x01(\x0e\x32\'.cz.proto.ingestion.BulkLoadStreamState\x12\x12\n\nsql_job_id\x18\x03 \x01(\t\x12\x37\n\nidentifier\x18\x04 \x01(\x0b\x32#.cz.proto.ingestion.TableIdentifier\x12>\n\toperation\x18\x05 \x01(\x0e\x32+.cz.proto.ingestion.BulkLoadStreamOperation\x12\x16\n\x0epartition_spec\x18\x06 \x01(\t\x12\x13\n\x0brecord_keys\x18\x07 \x03(\t\x12\x37\n\rstream_schema\x18\x08 \x01(\x0b\x32 .cz.proto.ingestion.StreamSchema\x12\x15\n\rsql_error_msg\x18\t \x01(\t\x12 \n\x18prefer_internal_endpoint\x18\n \x01(\x08\x12Y\n\x12\x65ncryption_options\x18\x0b \x03(\x0b\x32=.cz.proto.ingestion.BulkLoadStreamInfo.EncryptionOptionsEntry\x1a\x38\n\x16\x45ncryptionOptionsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xc9\x01\n\x1a\x42ulkLoadStreamWriterConfig\x12\x39\n\x0cstaging_path\x18\x01 \x01(\x0b\x32#.cz.proto.ingestion.StagingPathInfo\x12-\n\x0b\x66ile_format\x18\x02 \x01(\x0e\x32\x18.cz.proto.FileFormatType\x12\x1d\n\x15max_num_rows_per_file\x18\x03 \x01(\x03\x12\"\n\x1amax_size_in_bytes_per_file\x18\x04 \x01(\x03\"\x93\x02\n\x1b\x43reateBulkLoadStreamRequest\x12,\n\x07\x61\x63\x63ount\x18\x01 \x01(\x0b\x32\x1b.cz.proto.ingestion.Account\x12\x37\n\nidentifier\x18\x02 \x01(\x0b\x32#.cz.proto.ingestion.TableIdentifier\x12>\n\toperation\x18\x03 \x01(\x0e\x32+.cz.proto.ingestion.BulkLoadStreamOperation\x12\x16\n\x0epartition_spec\x18\x04 \x01(\t\x12\x13\n\x0brecord_keys\x18\x05 \x03(\t\x12 \n\x18prefer_internal_endpoint\x18\x06 \x01(\x08\"\x88\x01\n\x1c\x43reateBulkLoadStreamResponse\x12\x32\n\x06status\x18\x01 \x01(\x0b\x32\".cz.proto.ingestion.ResponseStatus\x12\x34\n\x04info\x18\x02 \x01(\x0b\x32&.cz.proto.ingestion.BulkLoadStreamInfo\"\xad\x01\n\x18GetBulkLoadStreamRequest\x12,\n\x07\x61\x63\x63ount\x18\x01 \x01(\x0b\x32\x1b.cz.proto.ingestion.Account\x12\x37\n\nidentifier\x18\x02 \x01(\x0b\x32#.cz.proto.ingestion.TableIdentifier\x12\x11\n\tstream_id\x18\x03 \x01(\t\x12\x17\n\x0fneed_table_meta\x18\x04 \x01(\x08\"\x85\x01\n\x19GetBulkLoadStreamResponse\x12\x32\n\x06status\x18\x01 \x01(\x0b\x32\".cz.proto.ingestion.ResponseStatus\x12\x34\n\x04info\x18\x02 \x01(\x0b\x32&.cz.proto.ingestion.BulkLoadStreamInfo\"\xeb\x02\n\x1b\x43ommitBulkLoadStreamRequest\x12,\n\x07\x61\x63\x63ount\x18\x01 \x01(\x0b\x32\x1b.cz.proto.ingestion.Account\x12\x37\n\nidentifier\x18\x02 \x01(\x0b\x32#.cz.proto.ingestion.TableIdentifier\x12\x11\n\tstream_id\x18\x03 \x01(\t\x12\x19\n\x11\x65xecute_workspace\x18\x04 \x01(\t\x12\x17\n\x0f\x65xecute_vc_name\x18\x05 \x01(\t\x12O\n\x0b\x63ommit_mode\x18\x06 \x01(\x0e\x32:.cz.proto.ingestion.CommitBulkLoadStreamRequest.CommitMode\x12\x1a\n\x12spec_partition_ids\x18\x07 \x03(\r\"1\n\nCommitMode\x12\x11\n\rCOMMIT_STREAM\x10\x00\x12\x10\n\x0c\x41\x42ORT_STREAM\x10\x01\"\x88\x01\n\x1c\x43ommitBulkLoadStreamResponse\x12\x32\n\x06status\x18\x01 \x01(\x0b\x32\".cz.proto.ingestion.ResponseStatus\x12\x34\n\x04info\x18\x02 \x01(\x0b\x32&.cz.proto.ingestion.BulkLoadStreamInfo\"\xb1\x01\n\x1fOpenBulkLoadStreamWriterRequest\x12,\n\x07\x61\x63\x63ount\x18\x01 \x01(\x0b\x32\x1b.cz.proto.ingestion.Account\x12\x37\n\nidentifier\x18\x02 \x01(\x0b\x32#.cz.proto.ingestion.TableIdentifier\x12\x11\n\tstream_id\x18\x03 \x01(\t\x12\x14\n\x0cpartition_id\x18\x04 \x01(\r\"\x96\x01\n OpenBulkLoadStreamWriterResponse\x12\x32\n\x06status\x18\x01 \x01(\x0b\x32\".cz.proto.ingestion.ResponseStatus\x12>\n\x06\x63onfig\x18\x02 \x01(\x0b\x32..cz.proto.ingestion.BulkLoadStreamWriterConfig\"\xe3\x01\n!FinishBulkLoadStreamWriterRequest\x12,\n\x07\x61\x63\x63ount\x18\x01 \x01(\x0b\x32\x1b.cz.proto.ingestion.Account\x12\x37\n\nidentifier\x18\x02 \x01(\x0b\x32#.cz.proto.ingestion.TableIdentifier\x12\x11\n\tstream_id\x18\x03 \x01(\t\x12\x14\n\x0cpartition_id\x18\x04 \x01(\r\x12\x15\n\rwritten_files\x18\x05 \x03(\t\x12\x17\n\x0fwritten_lengths\x18\x06 \x03(\x04\"X\n\"FinishBulkLoadStreamWriterResponse\x12\x32\n\x06status\x18\x01 \x01(\x0b\x32\".cz.proto.ingestion.ResponseStatus\"\x9c\x01\n\x16GetRouteWorkersRequest\x12\x38\n\x0btable_ident\x18\x01 \x01(\x0b\x32#.cz.proto.ingestion.TableIdentifier\x12\x35\n\x0c\x63onnect_mode\x18\x02 \x01(\x0e\x32\x1f.cz.proto.ingestion.ConnectMode\x12\x11\n\ttablet_id\x18\x03 \x03(\x03\"+\n\rHostPortTuple\x12\x0c\n\x04host\x18\x01 \x01(\t\x12\x0c\n\x04port\x18\x02 \x01(\x05\"\x92\x01\n\x17GetRouteWorkersResponse\x12\x30\n\x05tuple\x18\x01 \x03(\x0b\x32!.cz.proto.ingestion.HostPortTuple\x12\x11\n\ttablet_id\x18\x02 \x03(\x03\x12\x32\n\x06status\x18\x03 \x01(\x0b\x32\".cz.proto.ingestion.ResponseStatus\"H\n\x11OperationTypeList\x12\x33\n\x08op_types\x18\x01 \x03(\x0e\x32!.cz.proto.ingestion.OperationType\"\xf6\x01\n\tDataBlock\x12\x15\n\rarrow_payload\x18\x01 \x01(\x0c\x12\x1e\n\x16is_set_bitmaps_payload\x18\x02 \x01(\x0c\x12\x41\n\x10row_op_type_list\x18\x03 \x01(\x0b\x32%.cz.proto.ingestion.OperationTypeListH\x00\x12:\n\rblock_op_type\x18\x04 \x01(\x0e\x32!.cz.proto.ingestion.OperationTypeH\x00\x12\x11\n\tbucket_id\x18\x05 \x01(\r\x12\x10\n\x08num_rows\x18\x06 \x01(\rB\x0e\n\x0cop_type_info\"c\n\x0fMutateRowStatus\x12&\n\x04\x63ode\x18\x01 \x01(\x0e\x32\x18.cz.proto.ingestion.Code\x12\x15\n\rerror_message\x18\x02 \x01(\t\x12\x11\n\trow_index\x18\x03 \x01(\x05\"\xd6\x01\n\rMutateRequest\x12,\n\x07\x61\x63\x63ount\x18\x01 \x01(\x0b\x32\x1b.cz.proto.ingestion.Account\x12\x38\n\x0btable_ident\x18\x02 \x01(\x0b\x32#.cz.proto.ingestion.TableIdentifier\x12\x10\n\x08\x62\x61tch_id\x18\x03 \x01(\x03\x12\x17\n\x0fwrite_timestamp\x18\x04 \x01(\x03\x12\x32\n\x0b\x64\x61ta_blocks\x18\x05 \x03(\x0b\x32\x1d.cz.proto.ingestion.DataBlock\"\xa6\x01\n\x0eMutateResponse\x12\x10\n\x08\x62\x61tch_id\x18\x01 \x01(\x03\x12\x10\n\x08num_rows\x18\x02 \x01(\x03\x12\x32\n\x06status\x18\x03 \x01(\x0b\x32\".cz.proto.ingestion.ResponseStatus\x12<\n\x0frow_status_list\x18\x04 \x03(\x0b\x32#.cz.proto.ingestion.MutateRowStatus*\x91\x05\n\nMethodEnum\x12\x14\n\x10GATEWAY_RPC_CALL\x10\x00\x12\x12\n\x0eGET_TABLE_META\x10\x01\x12\x11\n\rCREATE_TABLET\x10\x02\x12\x15\n\x11GET_MUTATE_WORKER\x10\x03\x12\x11\n\rCOMMIT_TABLET\x10\x04\x12\x0f\n\x0b\x44ROP_TABLET\x10\x05\x12\x16\n\x12\x43HECK_TABLE_EXISTS\x10\x06\x12\x19\n\x15\x43REATE_PENDING_STREAM\x10\x07\x12\x19\n\x15\x43OMMIT_PENDING_STREAM\x10\x08\x12\x16\n\x12GET_PENDING_STREAM\x10\t\x12\x17\n\x13JOIN_PENDING_STREAM\x10\n\x12\x1b\n\x17\x43REATE_BULK_LOAD_STREAM\x10\x0b\x12\x18\n\x14GET_BULK_LOAD_STREAM\x10\x0c\x12\x1b\n\x17\x43OMMIT_BULK_LOAD_STREAM\x10\r\x12 \n\x1cOPEN_BULK_LOAD_STREAM_WRITER\x10\x0e\x12\"\n\x1e\x46INISH_BULK_LOAD_STREAM_WRITER\x10\x0f\x12\x1b\n\x17\x43REATE_OR_GET_STREAM_V2\x10\x10\x12\x13\n\x0f\x43LOSE_STREAM_V2\x10\x11\x12\x17\n\x13GET_ROUTE_WORKER_V2\x10\x12\x12\x1e\n\x1a\x43REATE_BULK_LOAD_STREAM_V2\x10\x13\x12\x1b\n\x17GET_BULK_LOAD_STREAM_V2\x10\x14\x12\x1e\n\x1a\x43OMMIT_BULK_LOAD_STREAM_V2\x10\x15\x12#\n\x1fOPEN_BULK_LOAD_STREAM_WRITER_V2\x10\x16\x12%\n!FINISH_BULK_LOAD_STREAM_WRITER_V2\x10\x17*1\n\x0cIGSTableType\x12\n\n\x06NORMAL\x10\x00\x12\x0b\n\x07\x43LUSTER\x10\x01\x12\x08\n\x04\x41\x43ID\x10\x02*.\n\x04\x43ode\x12\x0b\n\x07SUCCESS\x10\x00\x12\n\n\x06\x46\x41ILED\x10\x01\x12\r\n\tTHROTTLED\x10\x02*\x8c\x01\n\x13\x42ulkLoadStreamState\x12\x0e\n\nBL_CREATED\x10\x00\x12\r\n\tBL_SEALED\x10\x01\x12\x17\n\x13\x42L_COMMIT_SUBMITTED\x10\x02\x12\x15\n\x11\x42L_COMMIT_SUCCESS\x10\x03\x12\x14\n\x10\x42L_COMMIT_FAILED\x10\x04\x12\x10\n\x0c\x42L_CANCELLED\x10\x05*I\n\x17\x42ulkLoadStreamOperation\x12\r\n\tBL_APPEND\x10\x00\x12\x10\n\x0c\x42L_OVERWRITE\x10\x01\x12\r\n\tBL_UPSERT\x10\x02*<\n\x0b\x43onnectMode\x12\n\n\x06\x44IRECT\x10\x00\x12\x0b\n\x07GATEWAY\x10\x01\x12\x14\n\x10GATEWAY_INTERNAL\x10\x02*\x85\x01\n\rOperationType\x12\x0b\n\x07UNKNOWN\x10\x00\x12\n\n\x06INSERT\x10\x01\x12\n\n\x06UPDATE\x10\x02\x12\n\n\x06\x44\x45LETE\x10\x03\x12\n\n\x06UPSERT\x10\x04\x12\x11\n\rINSERT_IGNORE\x10\x05\x12\x11\n\rUPDATE_IGNORE\x10\x06\x12\x11\n\rDELETE_IGNORE\x10\x07\x32\xb3\x08\n\x1aIngestionControllerService\x12Y\n\x0eGatewayRpcCall\x12\".cz.proto.ingestion.GatewayRequest\x1a#.cz.proto.ingestion.GatewayResponse\x12p\n\x11\x43reateOrGetStream\x12,.cz.proto.ingestion.CreateOrGetStreamRequest\x1a-.cz.proto.ingestion.CreateOrGetStreamResponse\x12^\n\x0b\x43loseStream\x12&.cz.proto.ingestion.CloseStreamRequest\x1a\'.cz.proto.ingestion.CloseStreamResponse\x12j\n\x0fGetRouteWorkers\x12*.cz.proto.ingestion.GetRouteWorkersRequest\x1a+.cz.proto.ingestion.GetRouteWorkersResponse\x12y\n\x14\x43reateBulkLoadStream\x12/.cz.proto.ingestion.CreateBulkLoadStreamRequest\x1a\x30.cz.proto.ingestion.CreateBulkLoadStreamResponse\x12p\n\x11GetBulkLoadStream\x12,.cz.proto.ingestion.GetBulkLoadStreamRequest\x1a-.cz.proto.ingestion.GetBulkLoadStreamResponse\x12y\n\x14\x43ommitBulkLoadStream\x12/.cz.proto.ingestion.CommitBulkLoadStreamRequest\x1a\x30.cz.proto.ingestion.CommitBulkLoadStreamResponse\x12\x85\x01\n\x18OpenBulkLoadStreamWriter\x12\x33.cz.proto.ingestion.OpenBulkLoadStreamWriterRequest\x1a\x34.cz.proto.ingestion.OpenBulkLoadStreamWriterResponse\x12\x8b\x01\n\x1a\x46inishBulkLoadStreamWriter\x12\x35.cz.proto.ingestion.FinishBulkLoadStreamWriterRequest\x1a\x36.cz.proto.ingestion.FinishBulkLoadStreamWriterResponse2\xca\x01\n\x16IngestionWorkerService\x12S\n\x06Mutate\x12!.cz.proto.ingestion.MutateRequest\x1a\".cz.proto.ingestion.MutateResponse(\x01\x30\x01\x12[\n\x0eMutateInternal\x12!.cz.proto.ingestion.MutateRequest\x1a\".cz.proto.ingestion.MutateResponse(\x01\x30\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ingestion_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_BULKLOADSTREAMINFO_ENCRYPTIONOPTIONSENTRY']._loaded_options = None
  _globals['_BULKLOADSTREAMINFO_ENCRYPTIONOPTIONSENTRY']._serialized_options = b'8\001'
  _globals['_METHODENUM']._serialized_start=6987
  _globals['_METHODENUM']._serialized_end=7644
  _globals['_IGSTABLETYPE']._serialized_start=7646
  _globals['_IGSTABLETYPE']._serialized_end=7695
  _globals['_CODE']._serialized_start=7697
  _globals['_CODE']._serialized_end=7743
  _globals['_BULKLOADSTREAMSTATE']._serialized_start=7746
  _globals['_BULKLOADSTREAMSTATE']._serialized_end=7886
  _globals['_BULKLOADSTREAMOPERATION']._serialized_start=7888
  _globals['_BULKLOADSTREAMOPERATION']._serialized_end=7961
  _globals['_CONNECTMODE']._serialized_start=7963
  _globals['_CONNECTMODE']._serialized_end=8023
  _globals['_OPERATIONTYPE']._serialized_start=8026
  _globals['_OPERATIONTYPE']._serialized_end=8159
  _globals['_GATEWAYREQUEST']._serialized_start=80
  _globals['_GATEWAYREQUEST']._serialized_end=174
  _globals['_GATEWAYRESPONSE']._serialized_start=176
  _globals['_GATEWAYRESPONSE']._serialized_end=269
  _globals['_USERIDENTIFIER']._serialized_start=271
  _globals['_USERIDENTIFIER']._serialized_end=363
  _globals['_ACCOUNT']._serialized_start=365
  _globals['_ACCOUNT']._serialized_end=445
  _globals['_TABLEIDENTIFIER']._serialized_start=447
  _globals['_TABLEIDENTIFIER']._serialized_end=545
  _globals['_GATEWAYRESPONSESTATUS']._serialized_start=547
  _globals['_GATEWAYRESPONSESTATUS']._serialized_end=647
  _globals['_RESPONSESTATUS']._serialized_start=649
  _globals['_RESPONSESTATUS']._serialized_end=748
  _globals['_DATAFIELD']._serialized_start=750
  _globals['_DATAFIELD']._serialized_end=809
  _globals['_DISTRIBUTIONSPEC']._serialized_start=811
  _globals['_DISTRIBUTIONSPEC']._serialized_end=893
  _globals['_PRIMARYKEYSPEC']._serialized_start=895
  _globals['_PRIMARYKEYSPEC']._serialized_end=930
  _globals['_PARTITIONSPEC']._serialized_start=932
  _globals['_PARTITIONSPEC']._serialized_end=970
  _globals['_STREAMSCHEMA']._serialized_start=973
  _globals['_STREAMSCHEMA']._serialized_end=1217
  _globals['_CREATEORGETSTREAMREQUEST']._serialized_start=1220
  _globals['_CREATEORGETSTREAMREQUEST']._serialized_end=1371
  _globals['_CREATEORGETSTREAMRESPONSE']._serialized_start=1374
  _globals['_CREATEORGETSTREAMRESPONSE']._serialized_end=1590
  _globals['_CLOSESTREAMREQUEST']._serialized_start=1592
  _globals['_CLOSESTREAMREQUEST']._serialized_end=1716
  _globals['_CLOSESTREAMRESPONSE']._serialized_start=1718
  _globals['_CLOSESTREAMRESPONSE']._serialized_end=1791
  _globals['_OSSSTAGINGPATHINFO']._serialized_start=1794
  _globals['_OSSSTAGINGPATHINFO']._serialized_end=1967
  _globals['_COSSTAGINGPATHINFO']._serialized_start=1970
  _globals['_COSSTAGINGPATHINFO']._serialized_end=2110
  _globals['_S3STAGINGPATHINFO']._serialized_start=2113
  _globals['_S3STAGINGPATHINFO']._serialized_end=2250
  _globals['_GCSSTAGINGPATHINFO']._serialized_start=2253
  _globals['_GCSSTAGINGPATHINFO']._serialized_end=2393
  _globals['_OBSSTAGINGPATHINFO']._serialized_start=2396
  _globals['_OBSSTAGINGPATHINFO']._serialized_end=2538
  _globals['_TOSSTAGINGPATHINFO']._serialized_start=2541
  _globals['_TOSSTAGINGPATHINFO']._serialized_end=2683
  _globals['_LOCALSTAGINGPATHINFO']._serialized_start=2685
  _globals['_LOCALSTAGINGPATHINFO']._serialized_end=2721
  _globals['_STAGINGPATHINFO']._serialized_start=2724
  _globals['_STAGINGPATHINFO']._serialized_end=3176
  _globals['_BULKLOADSTREAMINFO']._serialized_start=3179
  _globals['_BULKLOADSTREAMINFO']._serialized_end=3730
  _globals['_BULKLOADSTREAMINFO_ENCRYPTIONOPTIONSENTRY']._serialized_start=3674
  _globals['_BULKLOADSTREAMINFO_ENCRYPTIONOPTIONSENTRY']._serialized_end=3730
  _globals['_BULKLOADSTREAMWRITERCONFIG']._serialized_start=3733
  _globals['_BULKLOADSTREAMWRITERCONFIG']._serialized_end=3934
  _globals['_CREATEBULKLOADSTREAMREQUEST']._serialized_start=3937
  _globals['_CREATEBULKLOADSTREAMREQUEST']._serialized_end=4212
  _globals['_CREATEBULKLOADSTREAMRESPONSE']._serialized_start=4215
  _globals['_CREATEBULKLOADSTREAMRESPONSE']._serialized_end=4351
  _globals['_GETBULKLOADSTREAMREQUEST']._serialized_start=4354
  _globals['_GETBULKLOADSTREAMREQUEST']._serialized_end=4527
  _globals['_GETBULKLOADSTREAMRESPONSE']._serialized_start=4530
  _globals['_GETBULKLOADSTREAMRESPONSE']._serialized_end=4663
  _globals['_COMMITBULKLOADSTREAMREQUEST']._serialized_start=4666
  _globals['_COMMITBULKLOADSTREAMREQUEST']._serialized_end=5029
  _globals['_COMMITBULKLOADSTREAMREQUEST_COMMITMODE']._serialized_start=4980
  _globals['_COMMITBULKLOADSTREAMREQUEST_COMMITMODE']._serialized_end=5029
  _globals['_COMMITBULKLOADSTREAMRESPONSE']._serialized_start=5032
  _globals['_COMMITBULKLOADSTREAMRESPONSE']._serialized_end=5168
  _globals['_OPENBULKLOADSTREAMWRITERREQUEST']._serialized_start=5171
  _globals['_OPENBULKLOADSTREAMWRITERREQUEST']._serialized_end=5348
  _globals['_OPENBULKLOADSTREAMWRITERRESPONSE']._serialized_start=5351
  _globals['_OPENBULKLOADSTREAMWRITERRESPONSE']._serialized_end=5501
  _globals['_FINISHBULKLOADSTREAMWRITERREQUEST']._serialized_start=5504
  _globals['_FINISHBULKLOADSTREAMWRITERREQUEST']._serialized_end=5731
  _globals['_FINISHBULKLOADSTREAMWRITERRESPONSE']._serialized_start=5733
  _globals['_FINISHBULKLOADSTREAMWRITERRESPONSE']._serialized_end=5821
  _globals['_GETROUTEWORKERSREQUEST']._serialized_start=5824
  _globals['_GETROUTEWORKERSREQUEST']._serialized_end=5980
  _globals['_HOSTPORTTUPLE']._serialized_start=5982
  _globals['_HOSTPORTTUPLE']._serialized_end=6025
  _globals['_GETROUTEWORKERSRESPONSE']._serialized_start=6028
  _globals['_GETROUTEWORKERSRESPONSE']._serialized_end=6174
  _globals['_OPERATIONTYPELIST']._serialized_start=6176
  _globals['_OPERATIONTYPELIST']._serialized_end=6248
  _globals['_DATABLOCK']._serialized_start=6251
  _globals['_DATABLOCK']._serialized_end=6497
  _globals['_MUTATEROWSTATUS']._serialized_start=6499
  _globals['_MUTATEROWSTATUS']._serialized_end=6598
  _globals['_MUTATEREQUEST']._serialized_start=6601
  _globals['_MUTATEREQUEST']._serialized_end=6815
  _globals['_MUTATERESPONSE']._serialized_start=6818
  _globals['_MUTATERESPONSE']._serialized_end=6984
  _globals['_INGESTIONCONTROLLERSERVICE']._serialized_start=8162
  _globals['_INGESTIONCONTROLLERSERVICE']._serialized_end=9237
  _globals['_INGESTIONWORKERSERVICE']._serialized_start=9240
  _globals['_INGESTIONWORKERSERVICE']._serialized_end=9442
# @@protoc_insertion_point(module_scope)



import datetime
import decimal
import string
import uuid

import os

import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
import pyarrow.compute as pc
from logging import getLogger
from qcloud_cos import CosConfig, CosServiceError, CosClientError
from qcloud_cos import CosS3Client

from clickzetta.proto.generated import ingestion_pb2, data_type_pb2, file_format_type_pb2
from clickzetta.client import Client
from clickzetta.bulkload.bulkload_enums import BulkLoadMetaData, BulkLoadConfig, BulkLoadState, FileFormatType

_logger = getLogger(__name__)


class Row:
    def __init__(self, full_fields: dict, table_name: string):
        self.field_name_values = {}
        self.full_fields = full_fields
        self.table_name = table_name

    def set_value(self, field_name: string, field_value):
        if field_name in self.full_fields:
            self.field_name_values[field_name] = field_value
        else:
            raise RuntimeError('Field name:{} is not in table:{}'.format(field_name, self.table_name))


class BulkLoadWriter:
    def __init__(self, client: Client, meta_data: BulkLoadMetaData, config: BulkLoadConfig, partition_id: int):
        self.client = client
        self.meta_data = meta_data
        self.config = config
        self.partition_id = partition_id
        self.staging_config = config.get_staging_config(meta_data.get_prefer_internal_endpoint())
        self._process_staging_type()
        self.file_io = config.get_staging_config(meta_data.get_prefer_internal_endpoint()).create_file_io()
        self.file_format = config.get_file_format()
        self.max_file_records = config.get_max_rows_per_file()
        self.max_file_size = config.get_max_file_size_per_file()
        if meta_data.get_state() != BulkLoadState.CREATED:
            raise AssertionError("Failed to create BulkLoadStream due to invalid state:" + meta_data.get_state())
        self.partition_spec = self._parse_partition_spec(self.meta_data.table.schema,
                                                         self.meta_data.get_partition_specs())
        self.finished_files = []
        self.finished_file_sizes = []
        self.file_name_uuid = uuid.uuid4()
        self.file_id = 0
        self.closed = False
        self.current_total_rows = 0
        self.current_total_size = 0
        self.current_record_batch = {}
        self.current_batch_rows = 0
        # currently buffered size
        self.current_buffer_size = 0
        self.estimate_row_static_size = self._estimate_row_static_size()
        self.pyarrow_schema = self._generate_pyarrow_schema(self.meta_data.get_table().schema)
        self.writer = None

    def _process_staging_type(self):
        if self.staging_config.type == 'oss':
            self.location = self.staging_config.path
        elif self.staging_config.type == 'cos':
            self.location = '/tmp/' + self.staging_config.path
            cos_config = CosConfig(Region=self.staging_config.endpoint, SecretId=self.staging_config.id,
                                   SecretKey=self.staging_config.secret, Token=self.staging_config.token)
            self.cos_client = CosS3Client(cos_config)
        elif self.staging_config.type == 's3':
            self.location = self.staging_config.path
        elif self.staging_config.type == 'gcs':
            self.location = self.staging_config.path
        else:
            self.location = self.staging_config.path

    def get_stream_id(self):
        return self.meta_data.get_stream_id()

    def get_operation(self):
        return self.meta_data.get_operation()

    def get_schema(self):
        return self.meta_data.get_table().schema

    def get_table(self):
        return self.meta_data.get_table()

    def get_partition_id(self):
        return self.partition_id

    def create_row(self):
        return Row(self.meta_data.get_table().schema, self.meta_data.get_table_name())

    def write(self, row: Row):
        self._check_file_status()
        if not self.current_record_batch:
            self._construct_new_record_batch()

        for partition in self.partition_spec:
            row.field_name_values[partition] = self.partition_spec[partition]
        current_row_size = self.estimate_row_static_size
        for column in self.current_record_batch.keys():
            if column not in row.field_name_values:
                row.field_name_values[column] = None
        for filed_name in row.field_name_values:
            field_value = row.field_name_values[filed_name]
            self.current_record_batch[filed_name].append(field_value)
            if isinstance(field_value, str):
                current_row_size += len(field_value)
        self.current_buffer_size = self.current_buffer_size + current_row_size
        self.current_batch_rows = self.current_batch_rows + 1
        self.current_total_rows = self.current_total_rows + 1

        # batch rows set to 40000 or 16MB(in arrow), will get from session config later
        if self.current_batch_rows == 40000 or self.current_buffer_size >= 16 * 1024 * 1024:
            buffer_size = self._flush_record_batch()
            self.current_total_size = self.current_total_size + buffer_size
            self.current_buffer_size = 0

    def finish(self):
        if self.closed:
            raise AssertionError('BulkLoadWriter is already closed.')
        self._close_current_file()
        status = self.client.finish_bulkload_stream_writer(self.meta_data.get_instance_id(),
                                                           self.meta_data.get_workspace(),
                                                           self.meta_data.get_schema_name(),
                                                           self.meta_data.get_table_name(),
                                                           self.meta_data.get_stream_id(), self.partition_id,
                                                           self.finished_files, self.finished_file_sizes)
        if status.code != ingestion_pb2.Code.SUCCESS:
            msg = status.message
            raise RuntimeError(
                'Finish BulkLoadStreamWriter for instance:{}, workspace:{}, '
                'table:{}.{}, streamId:{}, partitionId:{} failed. Error message:{}'.format(
                    self.meta_data.get_instance_id(),
                    self.meta_data.get_workspace(), self.meta_data.get_schema_name(), self.meta_data.get_table_name(),
                    self.meta_data.get_stream_id(), self.partition_id, msg))
        _logger.info("Flush bulk load stream {} partitionId {} with {} files".format(self.meta_data.get_stream_id(),
                                                                                     self.partition_id,
                                                                                     len(self.finished_files)))
        self.finished_files.clear()
        self.finished_file_sizes.clear()
        self.closed = True

    def abort(self):
        self.writer.close()
        self.current_total_rows = 0
        self.current_total_size = 0
        self.current_batch_rows = 0
        self.current_record_batch = {}
        self.file_io.close()
        self.closed = True

    def close(self):
        if self.closed:
            return
        self.finish()

    def _flush_record_batch(self):
        if self.current_batch_rows == 0:
            return 0
        batch_data = []
        for item in self.current_record_batch:
            batch_data.append(self._convert_data_type(item, self.current_record_batch[item]))
        batch = pa.record_batch(batch_data, schema=self.pyarrow_schema)
        self.writer.write_batch(batch)
        self.current_batch_rows = 0
        self.current_record_batch.clear()
        return batch.get_total_buffer_size()

    def _estimate_type_size(self, data_type: data_type_pb2.DataType):
        if data_type.category == data_type_pb2.DataTypeCategory.INT8:
            return 1
        elif data_type.category == data_type_pb2.DataTypeCategory.INT16:
            return 2
        elif data_type.category == data_type_pb2.DataTypeCategory.INT32:
            return 4
        elif data_type.category == data_type_pb2.DataTypeCategory.INT64:
            return 8
        elif data_type.category == data_type_pb2.DataTypeCategory.FLOAT32:
            return 4
        elif data_type.category == data_type_pb2.DataTypeCategory.FLOAT64:
            return 8
        elif data_type.category == data_type_pb2.DataTypeCategory.DECIMAL:
            return 16
        elif data_type.category == data_type_pb2.DataTypeCategory.BOOLEAN:
            return 1
        elif data_type.category == data_type_pb2.DataTypeCategory.CHAR or \
                data_type.category == data_type_pb2.DataTypeCategory.VARCHAR or \
                data_type.category == data_type_pb2.DataTypeCategory.STRING:
            return 16
        elif data_type.category == data_type_pb2.DataTypeCategory.DATE:
            return 4
        elif data_type.category == data_type_pb2.DataTypeCategory.TIMESTAMP_LTZ:
            return 8
        # as default
        return 8

    def _estimate_row_static_size(self):
        schema = self.get_schema()
        row_size = 0
        for column_name, column_type in schema.items():
            row_size += self._estimate_type_size(column_type)
        return row_size

    def _convert_data_type(self, field_name: str, field_data: list):
        """Use common utility to convert input data to arrow array with correct arrow data type"""
        schema = self.meta_data.get_table().schema
        data_type = schema[field_name]
        arrow_type = to_arrow_type(data_type)

        if not field_data:
            return pa.array([], type=arrow_type)

        # Convert None, NaT, NaN to None
        batch_data = []
        for value in field_data:
            if value is None:
                batch_data.append(None)
            else:
                # Skip isna check for iterables (except strings)
                if not isinstance(value, (str, bytes)) and hasattr(value, '__iter__'):
                    batch_data.append(value)
                else:
                    try:
                        is_na = pd.isna(value)
                        # Handle scalar values
                        if isinstance(is_na, bool):
                            batch_data.append(None if is_na else value)
                        # Handle array-like values - just append as is
                        else:
                            batch_data.append(value)
                    except:
                        # If isna fails, just use the value as is
                        batch_data.append(value)

        return pc.cast(batch_data, arrow_type)


    def _construct_new_record_batch(self):
        self.current_record_batch.clear()
        table_fields = self.meta_data.get_table().schema
        for filed in table_fields:
            self.current_record_batch[filed] = []
        self.current_batch_rows = 0

    def _generate_pyarrow_schema(self, schema: dict) -> pa.Schema:
        pyarrow_fields = []
        for field in schema:
            data_type = schema[field]
            if data_type.category == data_type_pb2.DataTypeCategory.INT8:
                pyarrow_fields.append(pa.field(field, pa.int8()))
            elif data_type.category == data_type_pb2.DataTypeCategory.INT16:
                pyarrow_fields.append(pa.field(field, pa.int16()))
            elif data_type.category == data_type_pb2.DataTypeCategory.INT32:
                pyarrow_fields.append(pa.field(field, pa.int32()))
            elif data_type.category == data_type_pb2.DataTypeCategory.INT64:
                pyarrow_fields.append(pa.field(field, pa.int64()))
            elif data_type.category == data_type_pb2.DataTypeCategory.FLOAT32:
                pyarrow_fields.append(pa.field(field, pa.float32()))
            elif data_type.category == data_type_pb2.DataTypeCategory.FLOAT64:
                pyarrow_fields.append(pa.field(field, pa.float64()))
            elif data_type.category == data_type_pb2.DataTypeCategory.DECIMAL:
                precision = data_type.decimalTypeInfo.precision
                scale = data_type.decimalTypeInfo.scale
                pyarrow_fields.append(pa.field(field, pa.decimal128(precision, scale)))
            elif data_type.category == data_type_pb2.DataTypeCategory.BOOLEAN:
                pyarrow_fields.append(pa.field(field, pa.bool_()))
            elif data_type.category == data_type_pb2.DataTypeCategory.CHAR or \
                    data_type.category == data_type_pb2.DataTypeCategory.VARCHAR or \
                    data_type.category == data_type_pb2.DataTypeCategory.STRING:
                pyarrow_fields.append(pa.field(field, pa.string()))
            elif data_type.category == data_type_pb2.DataTypeCategory.DATE:
                pyarrow_fields.append(pa.field(field, pa.date32()))
            elif data_type.category == data_type_pb2.DataTypeCategory.TIMESTAMP_LTZ:
                timestamp_unit = data_type.timestamp_info.tsUnit
                if timestamp_unit == data_type_pb2.TimestampUnit.SECONDS:
                    pyarrow_fields.append(pa.field(field, pa.timestamp('s', tz='UTC')))
                elif timestamp_unit == data_type_pb2.TimestampUnit.MILLISECONDS:
                    pyarrow_fields.append(pa.field(field, pa.timestamp('ms', tz='UTC')))
                elif timestamp_unit == data_type_pb2.TimestampUnit.MICROSECONDS:
                    pyarrow_fields.append(pa.field(field, pa.timestamp('us', tz='UTC')))
                elif timestamp_unit == data_type_pb2.TimestampUnit.NANOSECONDS:
                    pyarrow_fields.append(pa.field(field, pa.timestamp('ns', tz='UTC')))
        return pa.schema(pyarrow_fields)

    def _parse_partition_spec(self, schema: dict, partition_spec: string) -> dict:
        partition_value_dict = {}
        if partition_spec.strip() == '':
            return partition_value_dict
        partition_pairs = partition_spec.strip().split(',')
        for partition in partition_pairs:
            kv = partition.strip().split('=')
            partition_value_dict[kv[0].strip()] = kv[1].strip()
        for field in schema:
            if field in partition_value_dict:
                data_type = schema[field]
                if data_type.category == data_type_pb2.DataTypeCategory.INT8 or \
                        data_type.category == data_type_pb2.DataTypeCategory.INT16 or \
                        data_type.category == data_type_pb2.DataTypeCategory.INT32:
                    partition_value_dict[field] = int(partition_value_dict[field])
                elif data_type.category == data_type_pb2.DataTypeCategory.INT64:
                    partition_value_dict[field] = int(partition_value_dict[field])
                elif data_type.category == data_type_pb2.DataTypeCategory.FLOAT32 or \
                        data_type.category == data_type_pb2.DataTypeCategory.FLOAT64:
                    partition_value_dict[field] = float(partition_value_dict[field])
                elif data_type.category == data_type_pb2.DataTypeCategory.DECIMAL:
                    partition_value_dict[field] = decimal.Decimal(partition_value_dict[field])
                elif data_type.category == data_type_pb2.DataTypeCategory.BOOLEAN:
                    partition_value_dict[field] = bool(partition_value_dict[field])
                elif data_type.category == data_type_pb2.DataTypeCategory.DATE:
                    partition_value_dict[field] = (
                            datetime.strptime(partition_value_dict[field], '%Y-%m-%d') - datetime(1970, 1, 1)).days

        return partition_value_dict

    def _get_current_file_name(self):
        return '{}{}-{}.{}'.format(self.location, self.file_name_uuid, self.file_id, self.file_format.value)

    def _create_next_file_writer(self):
        file_name = self._get_current_file_name()
        if self.staging_config.type == 'cos':
            if not os.path.exists(os.path.dirname(file_name)):
                self.file_io.create_dir(os.path.dirname(file_name))
        if self.file_format == FileFormatType.PARQUET:
            return pq.ParquetWriter(file_name, self.pyarrow_schema, filesystem=self.file_io)
        else:
            raise NotImplementedError('File format:{} is not supported yet.'.format(self.file_format.value))

    def _close_current_file(self):
        if self.writer is not None:
            buffer_size = self._flush_record_batch()
            self.writer.close()
            file_name = self._upload_local_file()
            self.finished_files.append(file_name)
            self.finished_file_sizes.append(self.current_total_size + buffer_size)
            self.current_total_size = 0
            self.current_total_rows = 0
            self.current_batch_rows = 0
            self.current_record_batch.clear()
            self.file_id = self.file_id + 1
            self.writer = None

    def _upload_local_file(self):
        if self.staging_config.type == 'cos':
            try:
                file_name = self._get_current_file_name()[5:]
                path_infos = file_name.split('/', 3)
                bucket = path_infos[2]
                key = path_infos[3]
                response = self.cos_client.upload_file(Bucket=bucket, LocalFilePath=self._get_current_file_name(),
                                                       Key=key, MAXThread=8, EnableMD5=True, PartSize=100)
                _logger.info('Upload file {} to cos successfully.Etag is {}'.format(file_name, response['ETag']))
                self.file_io.delete_file(self._get_current_file_name())
                _logger.info('Delete local file {} successfully.'.format(self._get_current_file_name()))
                return file_name
            except CosServiceError as e:
                raise e
            except CosClientError as e:
                raise e
            except Exception as e:
                raise e
        elif self.staging_config.type == 'oss':
            return self._get_current_file_name()
        elif self.staging_config.type == 's3':
            return self._get_current_file_name()
        elif self.staging_config.type == 'gcs':
            return self._get_current_file_name()
        else:
            return self._get_current_file_name()

    def _check_file_status(self):
        if self.writer is not None:
            if self.current_total_rows >= self.max_file_records or self.current_total_size > self.max_file_size:
                self._close_current_file()

        if self.writer is None:
            self.writer = self._create_next_file_writer()


def to_arrow_type(data_type: data_type_pb2.DataType) -> pa.DataType:
    """Convert ClickZetta data type to Arrow type"""
    category = data_type.category

    if category == data_type_pb2.DataTypeCategory.INT8:
        return pa.int8()
    elif category == data_type_pb2.DataTypeCategory.INT16:
        return pa.int16()
    elif category == data_type_pb2.DataTypeCategory.INT32:
        return pa.int32()
    elif category == data_type_pb2.DataTypeCategory.INT64:
        return pa.int64()
    elif category == data_type_pb2.DataTypeCategory.FLOAT32:
        return pa.float32()
    elif category == data_type_pb2.DataTypeCategory.FLOAT64:
        return pa.float64()
    elif category == data_type_pb2.DataTypeCategory.DECIMAL:
        if data_type.decimalTypeInfo:
            precision = data_type.decimalTypeInfo.precision
            scale = data_type.decimalTypeInfo.scale
            return pa.decimal128(precision, scale)
        return pa.decimal128(38, 10)  # Default precision and scale
    elif category == data_type_pb2.DataTypeCategory.BOOLEAN:
        return pa.bool_()
    elif category in (data_type_pb2.DataTypeCategory.VARCHAR,
                      data_type_pb2.DataTypeCategory.CHAR,
                      data_type_pb2.DataTypeCategory.STRING):
        return pa.string()
    elif category == data_type_pb2.DataTypeCategory.BINARY:
        return pa.binary()
    elif category == data_type_pb2.DataTypeCategory.DATE:
        return pa.date32()  # Days since Unix epoch
    elif category == data_type_pb2.DataTypeCategory.TIMESTAMP_LTZ:
        unit = _get_timestamp_unit(data_type.timestamp_info)
        return pa.timestamp(unit, tz='UTC')
    # TODO waiting for grpc proto update
    # elif category == data_type_pb2.DataTypeCategory.TIMESTAMP_NTZ:
    #     return pa.timestamp('us', tz=None)
    elif category == data_type_pb2.DataTypeCategory.INTERVAL_DAY_TIME:
        return pa.duration('us')
    elif category == data_type_pb2.DataTypeCategory.INTERVAL_YEAR_MONTH:
        return pa.duration('M')
    elif category == data_type_pb2.DataTypeCategory.ARRAY:
        element_type = to_arrow_type(data_type.arrayTypeInfo.elementType)
        return pa.list_(element_type)
    elif category == data_type_pb2.DataTypeCategory.MAP:
        key_type = to_arrow_type(data_type.mapTypeInfo.keyType)
        value_type = to_arrow_type(data_type.mapTypeInfo.valueType)
        return pa.map_(key_type, value_type)
    elif category == data_type_pb2.DataTypeCategory.STRUCT:
        fields = []
        for field in data_type.structTypeInfo.fields:
            field_type = to_arrow_type(field.type)
            fields.append(pa.field(field.name, field_type, field.type.nullable))
        return pa.struct(fields)
    # TODO waiting for grpc proto update
    # elif category == data_type_pb2.DataTypeCategory.VECTOR_TYPE:
    #     # Handle vector type as binary array
    #     vector_info = data_type.vector_info
    #     type_width = 0
    #
    #     if vector_info.numberType == data_type_pb2.VectorNumberType.I8:
    #         type_width = 1  # Byte.BYTES
    #     elif vector_info.numberType == data_type_pb2.VectorNumberType.I32:
    #         type_width = 4  # Integer.BYTES
    #     elif vector_info.numberType == data_type_pb2.VectorNumberType.F32:
    #         type_width = 4  # Float.BYTES
    #     else:
    #         raise ValueError(f"Unsupported vector number type: {vector_info.numberType}")
    #     # FixedSizeBinaryType
    #     return pa.binary(vector_info.dimension * type_width)
    #     # return pa.fixed_size_binary(vector_info.dimension * type_width)
    else:
        raise ValueError(f"Unsupported data type: {category}")


def _get_timestamp_unit(timestamp_info: data_type_pb2.TimestampInfo) -> str:
    """Get Arrow timestamp unit from proto timestamp info"""
    if not timestamp_info:
        return 'us'  # Default to microseconds

    unit = timestamp_info.tsUnit
    if unit == data_type_pb2.TimestampUnit.SECONDS:
        return 's'
    elif unit == data_type_pb2.TimestampUnit.MILLISECONDS:
        return 'ms'
    elif unit == data_type_pb2.TimestampUnit.MICROSECONDS:
        return 'us'
    elif unit == data_type_pb2.TimestampUnit.NANOSECONDS:
        return 'ns'
    return 'us'  # Default



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: file_format_type.proto
# Protobuf Python Version: 5.28.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    1,
    '',
    'file_format_type.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x16\x66ile_format_type.proto\x12\x08\x63z.proto*\x83\x01\n\x0e\x46ileFormatType\x12\x08\n\x04TEXT\x10\x00\x12\x0b\n\x07PARQUET\x10\x01\x12\x07\n\x03ORC\x10\x02\x12\x08\n\x04\x41VRO\x10\x03\x12\x07\n\x03\x43SV\x10\x04\x12\t\n\x05\x41RROW\x10\x05\x12\x0f\n\x0bHIVE_RESULT\x10`\x12\t\n\x05\x44UMMY\x10\x61\x12\n\n\x06MEMORY\x10\x62\x12\x0b\n\x07ICEBERG\x10\x63\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'file_format_type_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_FILEFORMATTYPE']._serialized_start=37
  _globals['_FILEFORMATTYPE']._serialized_end=168
# @@protoc_insertion_point(module_scope)
